from __future__ import annotations

import base64
import requests
import re
import wxcadm
from typing import Union, Optional
from wxcadm import log
from .common import *
from .common import _url_base
from .exceptions import *
from .cpapi import CPAPI
from .location import Location, LocationList
from .location_features import PagingGroup, VoicemailGroupList
from .auto_attendant import AutoAttendantList
from .call_queue import CallQueueList, OrgQueueSettings
from .hunt_group import HuntGroupList
from .webhooks import Webhooks
from .person import UserGroups, Person, PersonList
from .applications import WebexApplications
from .announcements import AnnouncementList, PlaylistList
from .workspace import Workspace, WorkspaceList, WorkspaceLocationList
from .call_routing import CallRouting, TranslationPatternList
from .reports import ReportList
from .calls import Calls
from .device import DeviceList, SupportedDeviceList
from .recording import ComplianceAnnouncementSettings, RecordingList, OrgRecordingVendorSelection
from .jobs import NumberManagementJobList, UserMoveJobList, RebuildPhonesJobList
from .virtual_line import VirtualLineList, VirtualLine
from .dect import DECTNetworkList
from .number import NumberList
from .events import AuditEventList
from .monitoring import MonitoringList
from .location_features import CallParkExtension


class Org:
    def __init__(self,
                 name: str,
                 id: str,
                 parent: wxcadm.Webex = None,
                 xsi: bool = False,
                 **kwargs
                 ):
        """Initialize an Org instance

        Args:
            name (str): The Organization name
            id (str): The Webex ID of the Organization
            parent (Webex, optional): The parent Webex instance that owns this Org.
            xsi (bool, optional): Whether to get the XSI Endpoints for the Org. Default False.

        Returns:
            Org: This instance of the Org class
        """

        # Instance attrs
        self._numbers = None
        self._paging_groups = None
        self._parent = parent
        self.name: str = name
        'The name of the Organization'
        self.id: str = id
        '''The Webex ID of the Organization'''
        self.xsi: dict = {}
        """The XSI details for the Organization"""
        self._params: dict = {"orgId": self.id}
        self._licenses: Optional[list] = None
        self._wxc_licenses: Optional[list] = None
        self._devices: Optional[list] = None
        self._usergroups: Optional[list] = None
        self._roles: Optional[dict] = None
        self._announcements: Optional[AnnouncementList] = None
        self._hunt_groups: Optional[HuntGroupList] = None
        self._call_queues: Optional[CallQueueList] = None
        self._workspaces: Optional[WorkspaceList] = None
        self._workspace_locations: Optional[WorkspaceLocationList] = None
        self._people: Optional[PersonList] = None
        self._auto_attendants: Optional[AutoAttendantList] = None
        self._locations: Optional[LocationList] = None
        self._compliance_announcement_settings: Optional[ComplianceAnnouncementSettings] = None
        self._number_management_jobs = None
        self._user_move_jobs = None
        self._rebuild_phones_jobs = None
        self._virtual_lines = None
        self._dect_networks = None
        self._voicemail_groups = None
        self._numbers = None
        self._supported_devices = None
        self._translation_patterns = None
        self._all_monitoring = None
        self._playlists = None

        self.call_routing = CallRouting(self)
        """ The :py:class:`CallRouting` instance for this Org """
        self._reports = None

        # Set the Authorization header based on how the instance was built
        self._headers = parent.headers

        # Create a CPAPI instance for CPAPI work
        self._cpapi = CPAPI(self, self._parent._access_token)

        if xsi:
            self.get_xsi_endpoints()


    @property
    def numbers(self):
        """ :class:`NumberList` of all numbers for the Org """
        if self._numbers is None:
            self._numbers = NumberList(self)
        return self._numbers

    @property
    def voicemail_groups(self):
        """ :class:`VoicemailGroupList of all Voicemail Groups for the Org """
        if self._voicemail_groups is None:
            self._voicemail_groups = VoicemailGroupList(self)
        return self._voicemail_groups

    @property
    def workspace_locations(self):
        """ :py:class:`WorkspaceLocationList` of all :py:class:`WorkspaceLocation`s"""
        if self._workspace_locations is None:
            self._workspace_locations = WorkspaceLocationList(self)
        return self._workspace_locations

    @property
    def reports(self) -> ReportList:
        if self._reports is None:
            self._reports = ReportList(self)
        return self._reports

    @property
    def spark_id(self):
        """ The decoded "Spark ID" of the Org ID"""
        org_id_bytes = base64.b64decode(self.id + "===")
        spark_id = org_id_bytes.decode("utf-8")
        return spark_id

    @property
    def org_id(self):
        """ The Org ID of the Org. :attr:`id` should be used in most cases. :attr:`org_id` just simplifies API calls """
        return self.id

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def compliance_announcement_settings(self):
        """ The Call Recording Compliance Announcement settings for the Org

        Returns:
            ComplianceAnnouncementSettings: The settings for the Org

        """
        if self._compliance_announcement_settings is None:
            response = webex_api_call('get', 'v1/telephony/config/callRecording/complianceAnnouncement',
                                      params={'orgId': self.id})
            self._compliance_announcement_settings = ComplianceAnnouncementSettings(self, **response)
        return self._compliance_announcement_settings

    @property
    def auto_attendants(self):
        if self._auto_attendants is None:
            self._auto_attendants = AutoAttendantList(self)
        return self._auto_attendants

    @property
    def workspaces(self):
        if self._workspaces is None:
            self._workspaces = WorkspaceList(parent=self)
        return self._workspaces

    @property
    def licenses(self):
        """ A list of all licenses for the Organization as a dictionary of names and IDs """
        if self._licenses is None:
            self._licenses = self.__get_licenses()
        return self._licenses

    @property
    def locations(self):
        if self._locations is None:
            self._locations = LocationList(self)
        return self._locations

    @property
    def dect_networks(self):
        if self._dect_networks is None:
            self._dect_networks = DECTNetworkList(self)
        return self._dect_networks

    @property
    def roles(self):
        """ A dict of user roles with the ID as the key and the Role name as the value """
        if self._roles is None:
            roles = {}
            response = webex_api_call('get', 'v1/roles', params={'orgId': self.id})
            for role in response:
                roles[role['id']] = role['name']
            self._roles = roles
        return self._roles

    @property
    def number_management_jobs(self):
        """ :class:`NumberManagementJobList` for this Organization """
        if self._number_management_jobs is None:
            self._number_management_jobs = NumberManagementJobList(self)
        return self._number_management_jobs

    @property
    def user_move_jobs(self):
        """ :class:`UserMoveJobList` for this Organization """
        if self._user_move_jobs is None:
            self._user_move_jobs = UserMoveJobList(self)
        return self._user_move_jobs

    @property
    def rebuild_phones_jobs(self):
        """ :class:`RebuildPhonesJobList` for this Organization """
        if self._rebuild_phones_jobs is None:
            self._rebuild_phones_jobs = RebuildPhonesJobList(self)
        return self._rebuild_phones_jobs

    @property
    def queue_settings(self):
        response = webex_api_call("get", "v1/telephony/config/queues/settings", params={'orgId': self.id})
        response['org'] = self
        return OrgQueueSettings.from_dict(response)

    def __get_licenses(self):
        """Gets all licenses for the Organization

        Returns:
            list: List of dictionaries containing the license name and ID

        """
        log.info("__get_licenses() started for org")
        license_list = []
        try:
            api_resp = webex_api_call("get", "v1/licenses", headers=self._headers, params=self._params)
        except APIError:
            return None
        for item in api_resp:
            if "Webex Calling" in item['name']:
                wxc_license = True
                if "Professional" in item['name']:
                    wxc_type = "person"
                elif "Workspace" in item['name']:
                    wxc_type = "workspace"
                elif "Basic" in item['name']:
                    wxc_type = "standard"
                else:
                    wxc_type = "unknown"
            else:
                wxc_license = False
                wxc_type = None
            lic = {"name": item['name'],
                   "id": item['id'],
                   "total": item['totalUnits'],
                   "consumed": item['consumedUnits'],
                   "subscription": item.get("subscriptionId", ""),
                   "wxc_license": wxc_license,
                   "wxc_type": wxc_type
                   }
            license_list.append(lic)
        return license_list

    @property
    def paging_groups(self):
        """ All the PagingGroups for the Org

        Returns:
            list[PagingGroup]: The PagingGroup instances for this Org

        """
        if not self._paging_groups:
            paging_groups = []
            response = webex_api_call("get", "v1/telephony/config/paging", headers=self._headers, params=self._params)
            for entry in response['locationPaging']:
                location = self.locations.get(id=entry['locationId'])
                this_pg = PagingGroup(location, entry['id'], entry['name'])
                paging_groups.append(this_pg)
            self._paging_groups = paging_groups
        return self._paging_groups

    @property
    def supported_devices(self) -> SupportedDeviceList:
        """ The :class:`SupportedDeviceList` of :class:`SupportedDevice` models for this Org """
        if self._supported_devices is None:
            self._supported_devices = SupportedDeviceList()
        return self._supported_devices

    @property
    def webhooks(self):
        """ The :py:class:`Webhooks` list with the :py:class:`Webhook` instances for the Org"""
        return Webhooks()

    @property
    def usergroups(self):
        """ The :py:class:`UserGroups` list with the :py:class:`UserGroup` instances for the Org """
        return UserGroups(parent=self)

    @property
    def applications(self):
        """ The :py:class:`WebexApplications` list with the :py:class:`WebexApplication` instances for this Org """
        return WebexApplications(parent=self)

    @property
    def virtual_lines(self):
        """ The :class:`VirtualLineList` list with the :class:`VirtualLine` instances for this Org """
        if self._virtual_lines is None:
            self._virtual_lines = VirtualLineList(self)
        return self._virtual_lines

    @property
    def announcements(self):
        """ The :py:class:`~wxcadm.announcements.AnnouncementList` list with the :py:class:`Announcement` instances
        for this Org"""
        if self._announcements is None:
            self._announcements = AnnouncementList(parent=self)
        return self._announcements

    @property
    def playlists(self) -> PlaylistList:
        """ The :py:class:`~wxcadm.playlists.PlaylistList` list with the :class:`Playlist` instances for this Org"""
        if self._playlists is None:
            self._playlists = PlaylistList(parent=self)
        return self._playlists

    @property
    def calls(self):
        """ The :py:class:`Calls` instance for this Org"""
        return Calls(parent=self)

    def get_paging_group(self, id: str = None, name: str = None, spark_id: str = None):
        """ Get the PagingGroup instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Paging
        Groups will be searched in order by ID, Name, and finally Spark ID. If no arguments are provided, the method
        will raise an Exception.

        Args:
            id (str, optional): The PagingGroup ID to find
            name (str, optional): The PagingGroup Name to find
            spark_id (str, optional): The Spark ID to find

        Returns:
            PagingGroup: The PagingGroup instance correlating to the given search argument. None is returned if no
            Location is found.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        for pg in self.paging_groups:
            if pg.id == id:
                return pg
        for pg in self.paging_groups:
            if pg.name == name:
                return pg
        for pg in self.paging_groups:
            if pg.spark_id == spark_id:
                return pg
        return None

    def get_number_assignment(self, number: str):
        """ Get the object to which a phone number is assigned.

        A number may be assigned to a Person, a Workspace, or any number of things. If the number is assigned to a
        known class, that instance will be returned. If not, the `owner` value from the API will be returned. If the
        number is found but not assigned, None will be returned.

        .. note::
            Since Webex sometimes uses E.164 formatting and other times uses the national format, the match is made
            using a partial match. If the method is passed a national-format number, but stored in Webex as an E164
            number, a match will be made.

        Args:
            number (str): The phone number to search for.

        """
        log.info(f"get_number_assignment({number})")
        for num in self.numbers:
            if num.phone_number is not None:
                if number in num.phone_number:
                    log.debug(f"Found match: {num}")
                    log.debug("Finding owner")
                    return num.owner
        return None

    def get_all_monitoring(self) -> dict:
        """ Returns a dict of all Users and Workspaces that are being monitored. The User (Person) or Workspace is the
        dict key and the Users and Workspaces that are monitoring that key are a list.

        Returns:
            dict: A dict in the format ``{ 'people': { person: [] }, 'workspaces': { person: [] } }``

        """
        if self._all_monitoring is None:
            all_monitoring = {'people': {}, 'workspaces': {}, 'park_extensions': {}, 'virtual_lines': {}}
            for person in self.people.webex_calling():
                monitoring: MonitoringList  = person.monitoring
                for element in monitoring.monitored_elements:
                    if isinstance(element, CallParkExtension):
                        if element.id not in all_monitoring['park_extensions'].keys():
                            all_monitoring['park_extensions'][element.id] = []
                        all_monitoring['park_extensions'][element.id].append(person)
                    if isinstance(element, VirtualLine):
                        if element.id not in all_monitoring['virtual_lines'].keys():
                            all_monitoring['virtual_lines'][element.id] = []
                        all_monitoring['virtual_lines'][element.id].append(person)
                    if isinstance(element, Person):
                        if element.id not in all_monitoring['people'].keys():
                            all_monitoring['people'][element.id] = []
                        all_monitoring['people'][element.id].append(person)
                    if isinstance(element, Workspace):
                        if element.id not in all_monitoring['workspaces'].keys():
                            all_monitoring['workspaces'][element.id] = []
                        all_monitoring['workspaces'][element.id].append(person)
            for workspace in self.workspaces.webex_calling():
                monitoring:monitoring = workspace.monitoring
                for element in monitoring.monitored_elements:
                    if isinstance(element, CallParkExtension):
                        if element.id not in all_monitoring['park_extensions'].keys():
                            all_monitoring['park_extensions'][element.id] = []
                        all_monitoring['park_extensions'][element.id].append(workspace)
                    if isinstance(element, VirtualLine):
                        if element.id not in all_monitoring['virtual_lines'].keys():
                            all_monitoring['virtual_lines'][element.id] = []
                        all_monitoring['virtual_lines'][element.id].append(workspace)
                    if isinstance(element, Person):
                        if element.id not in all_monitoring['people'].keys():
                            all_monitoring['people'][element.id] = []
                        all_monitoring['people'][element.id].append(workspace)
                    if isinstance(element, Workspace):
                        if element.id not in all_monitoring['workspaces'].keys():
                            all_monitoring['workspaces'][element.id] = []
                        all_monitoring['workspaces'][element.id].append(workspace)
            self._all_monitoring = all_monitoring
        return self._all_monitoring

    def get_workspace_devices(self, workspace: Optional[Workspace] = None):
        """ Get Webex Calling Workspaces and their associated Devices

        When called without an argument, a list of Workspace instances is returned. Devices associated with each
        Workspace can be accessed with the .devices attribute of each Workspace. If there are no Devices assigned
        to a Workspace, .devices will return an empty list (i.e. []).

        When called with a ``workspace`` argument, which is a Workspace instance, the Devices for that Workspace will
        be returned as a list of Device instances.

        Args:
            workspace (Workspace, optional): A Workspace instance to show devices for

        Returns:
            list: Either a list of :py:class:`Workspace` or a list of :py:class:`Device`, depending on the argument

        """
        # TODO: For now, until the Workspaces API works for all WxC Workspaces, we use the .numbers as the key
        if workspace is not None:
            return workspace.devices
        workspaces = []
        for number in self.numbers:
            if 'owner' in number.keys():
                if isinstance(number['owner'], dict) and number['owner']['type'] == 'PLACE':
                    print(number)
                    ws_config = {'displayName': number['owner']['firstName']}
                    workspace = Workspace(self, id=number['owner']['id'], config=ws_config)
                    if number['owner']['lastName'] != '.':
                        workspace.name += f" {number['owner']['lastName']}"
                    workspace.location = number['location']
                    workspaces.append(workspace)
        return workspaces

    @property
    def devices(self):
        """All the Device instances for the Org

        Returns:
            DeviceList: List of all Device instances

        """
        if self._devices is None:
            self._devices = DeviceList(self)
        return self._devices

    def get_device_by_id(self, device_id: str):
        """ Get the :py:class:`Device` instance for a specific Device ID

        Args:
            device_id (str): The Device ID to return

        Returns:
            Device: The Device instance. False is returned if no match is found.

        """
        log.debug(f"Finding device with ID {device_id}")
        for device in self.devices:
            if device.id == device_id or decode_spark_id(device.id).split("/")[-1] == \
                    decode_spark_id(device_id).split("/")[-1]:
                return device
        return False

    def get_location_by_name(self, name: str):
        """Get the Location instance associated with a given Location name

        .. deprecated:: 2.2.0
            Use :meth:`get_location` using the ```name``` argument instead.

        Args:
            name (str): The full name of the Location to look for. (Case sensitive)

        Returns:
            Location: The Location instance. If no match is found, None is returned

        """
        for location in self.locations:
            if location.name == name:
                return location
        return None

    @property
    def people(self):
        """ :py:class:`PersonList` wth all the :py:class:`Person` instances for the Organization """
        if self._people is None:
            self._people = PersonList(self)
        return self._people

    def get_person_by_id(self, id: str):
        """Get the Person instance associated with a given ID

        .. deprecated:: 4.0.0
            Use :py:meth:`Org.people.get_by_id()` instead

        Args:
            id (str): The Webex ID of the Person to look for.

        Returns:
            Person: The Person instance. If no match is found, None is returned

        """
        return self.people.get_by_id(id)

    @property
    def wxc_licenses(self):
        """Get only the Webex Calling licenses from the Org.licenses attribute

        Returns:
            list[str]: The license IDs for each Webex Calling license

        """
        if self.licenses is None:
            self.__get_licenses()
        license_list = []
        for license in self.licenses:
            if license['wxc_license']:
                license_list.append(license['id'])
        return license_list

    def get_wxc_person_license(self):
        """Get the Webex Calling - Professional license ID

        Returns:
            str: The License ID

        """
        log.info("__get_wxc_person_license started to find available Professional license")
        for license in self.licenses:
            if license['wxc_type'] == "person":
                return license['id']
        raise LicenseError("No Webex Calling Professional license found")

    def get_wxc_standard_license(self):
        """Get the Webex Calling - Basic license ID

        Returns:
            str: The License ID

        """
        log.info("__get_wxc_standard_license started to find available Basic license")
        for license in self.licenses:
            if license['wxc_type'] == "standard":
                return license['id']
        raise LicenseError("No Webex Calling Professional license found")

    def create_person(self, email: str,
                      location: Union[str, Location],
                      licenses: list = None,
                      calling: bool = True,
                      messaging: bool = True,
                      meetings: bool = True,
                      phone_number: str = None,
                      extension: str = None,
                      first_name: str = None,
                      last_name: str = None,
                      display_name: str = None,
                      ):
        """Create a new user in Webex.

        .. deprecated:: 4.0.0
            Use :py:meth:`Org.people.create()` instead.

        Args:
            email (str): The email address of the user
            location (Location): The Location instance to assign the user to. Also accepts the Location ID as a string
            licenses (list, optional): List of license IDs to assign to the user. Use this when the license IDs
                are known. To have the license IDs determined dynamically, use the `calling`, `messaging` and
                'meetings` parameters.
            calling (bool, optional): BETA - Whether to assign Calling licenses to the user. Defaults to True.
            messaging (bool, optional): BETA - Whether to assign Messaging licenses to the user. Defaults to True.
            meetings (bool, optional): BETA - Whether to assign Messaging licenses to the user. Defaults to True.
            phone_number (str, optional): The phone number to assign to the user.
            extension (str, optional): The extension to assign to the user
            first_name (str, optional): The user's first name. Defaults to empty string.
            last_name (str, optional): The users' last name. Defaults to empty string.
            display_name (str, optional): The full name of the user as displayed in Webex. If first name and last name
                are passed without display_name, the display name will be the concatenation of first and last name.

        Returns:
            Person: The Person instance of the newly-created user.

        """
        log.info(f"Creating new user: {email}")
        if (first_name or last_name) and display_name is None:
            log.debug("No display_name provided. Setting default.")
            display_name = f"{first_name} {last_name}"

        # Find the license IDs for each requested service, unless licenses was passed
        if not licenses:
            log.debug("No licenses specified. Finding licenses.")
            licenses = []
            if calling:
                log.debug("Calling requested. Finding Calling licenses.")
                log.debug(f"Licenses: {self.get_wxc_person_license()}")
                licenses.append(self.get_wxc_person_license())
            if messaging:
                pass
            if meetings:
                pass

        # Build the payload to send to the API
        log.debug("Building payload.")
        if isinstance(location, Location):
            location_id = location.id
        else:
            location_id = location
        payload = {"emails": [email], "locationId": location_id, "orgId": self.id, "licenses": licenses}
        if phone_number is not None:
            payload["phoneNumbers"] = [{"type": "work", "value": phone_number}]
        if extension is not None:
            payload["extension"] = extension
        if first_name is not None:
            payload["firstName"] = first_name
        if last_name is not None:
            payload["lastName"] = last_name
        if display_name is not None:
            payload["displayName"] = display_name
        log.debug(f"Payload: {payload}")
        r = requests.post(_url_base + "v1/people", headers=self._headers, params={"callingData": "true"},
                          json=payload)
        response = r.json()
        if r.status_code == 200:
            person = Person(response['id'], self, response)
            return person
        else:
            raise PutError(response['message'])

    def delete_person(self, person: Person):
        """ Delete a person

        This method deletes the person record from Webex, which removes all of their licenses, numbers, and devices.

        Args:
            person (Person): The :py:class:`Person` instance to delete

        Returns:
            bool: True on success, False otherwise

        """
        success = webex_api_call("delete", f"v1/people/{person.id}", params={'orgId': self.id})
        if success:
            self._people = []
            return True
        else:
            return False

    def get_person_by_email(self, email):
        """Get the Person instance from an email address

        .. deprecated:: 4.0.0
            Use :py:meth:`Org.people.get_by_email()` instead

        Args:
            email (str): The email of the Person to return

        Returns:
            Person: Person instance object. None in returned when no Person is found

        """
        return self.people.get_by_email(email)

    def get_xsi_endpoints(self):
        """Get the XSI endpoints for the Organization.

        Also stores them in the Org.xsi attribute.

        Returns:
            dict: Org.xsi attribute dictionary with each endpoint as an entry. None is returned if no XSI isn't enabled.

        """
        params = {"callingData": "true", **self._params}
        response = webex_api_call("get", "v1/organizations/" + self.id, headers=self._headers, params=params)
        if "xsiActionsEndpoint" in response:
            self.xsi['xsi_domain'] = response['xsiDomain']
            self.xsi['actions_endpoint'] = response['xsiActionsEndpoint']
            self.xsi['events_endpoint'] = response['xsiEventsEndpoint']
            self.xsi['events_channel_endpoint'] = response['xsiEventsChannelEndpoint']
        else:
            return None
        return self.xsi

    @property
    def call_queues(self) -> CallQueueList:
        """The Call Queues for an Organization.

        Returns:
            CallQueueList: List of CallQueue instances for the Organization

        """
        if self._call_queues is None:
            log.info("Getting Call Queues for Organization")
            self._call_queues = CallQueueList(self)
        return self._call_queues

    def get_call_queue_by_id(self, id: str):
        """ Get the :class:`CallQueue` instance with the requested ID

        .. deprecated:: 4.0.0
            Use :meth:`Org.call_queues.get()` instead

        Args:
            id (str): The CallQueue ID

        Returns:
            HuntGroup: The :class:`CallQueue` instance

        """
        log.info(f"Getting Call Queue with ID {id}")
        if not self._call_queues:
            self.call_queues
        for cq in self._call_queues:
            if cq.id == id:
                return cq
        return None

    def get_hunt_group_by_id(self, id: str):
        """ Get the :class:`HuntGroup` instance with the requested ID

        .. deprecated:: 4.0.0

            Use :meth:`Org.hunt_groups.get(id=)` instead


        Args:
            id (str): The HuntGroup ID

        Returns:
            HuntGroup: The :class:`HuntGroup` instance

        """
        if not self._hunt_groups:
            self.hunt_groups
        for hg in self._hunt_groups:
            if hg.id == id:
                return hg
        return None

    @property
    def hunt_groups(self) -> HuntGroupList:
        """The :class:`HuntGroupList` for the Organization.

        Returns:
            HuntGroupList: List of :class:`HuntGroup` instances for the Organization

        """
        if self._hunt_groups is None:
            log.info("Getting Hunt Groups for Organization")
            self._hunt_groups = HuntGroupList(self)
        return self._hunt_groups

    @property
    def wxc_people(self):
        """Return all the people within the Organization **who have Webex Calling**

        .. deprecated:: 4.0.0
            Use :py:meth:`Org.people.webex_calling()` instead

        Returns:
            list[Person]: List of Person instances of people who have a Webex Calling license

        """
        return self.people.webex_calling(True)

    @property
    def translation_patterns(self):
        """ The :class:`TranslationPatternList of Translation Patterns for the Org and all Locations """
        if self._translation_patterns is None:
            self._translation_patterns = TranslationPatternList(self)
        return self._translation_patterns

    @property
    def recording_vendor(self):
        return OrgRecordingVendorSelection(self)

    @property
    def recorded_people(self):
        """Return all the People within the Organization who have Call Recording enabled

        .. deprecated:: 4.0.0
            Use :py:meth:`Org.people.recorded()` instead

        Returns:
            list[Person]: List of Person instances that have Call Recording enabled

        """
        return self.people.recorded(True)

    def get_license_name(self, license_id: str):
        """Gets the name of a license by its ID

        Args:
            license_id (str): The License ID

        Returns:
            str: The License name. None if not found.

        """
        for license in self.licenses:
            if license['id'] == license_id:
                return license['name']
        return None

    def get_audit_events(self, start: str, end: str) -> AuditEventList:
        """ Get a list of Admin Audit Events for the Organization

        Args:
            start (str): The first date/time in the report, in the format `YYYY-MM-DD` or `YYYY-MM-DDTHH:MM:SS.000Z`
            end (str): The first date/time in the report, in the format `YYYY-MM-DD` or `YYYY-MM-DDTHH:MM:SS.000Z`

        Returns:
            AuditEventList: The :class:`AuditEventList` of all matching events

        """
        # Normalize the start and end if only the date was provided
        if re.fullmatch(r"\d{4}-\d{2}-\d{2}", start):
            start = start + "T00:00:00.000Z"
        if re.fullmatch(r"\d{4}-\d{2}-\d{2}", end):
            end = end + "T23:59:59.999Z"

        event_list = AuditEventList(self, start, end)
        return event_list

    def get_recordings(self, **kwargs):
        return RecordingList(parent=self, **kwargs)



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice, randint


class TestDectDataPull(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_dect_networks(self) -> None:
        org_dect_network = self.webex.org.dect_networks
        self.assertIsInstance(org_dect_network, wxcadm.dect.DECTNetworkList)
        if len(org_dect_network) > 0:
            self.assertIsInstance(org_dect_network[0], wxcadm.dect.DECTNetwork)
            with self.subTest('Location-based .dect_networks'):
                location = self.webex.org.locations.get(id=org_dect_network[0].location_id)
                self.assertIsInstance(location.dect_networks, wxcadm.dect.DECTNetworkList)
                self.assertIsInstance(location.dect_networks[0], wxcadm.dect.DECTNetwork)

    def test_dect_base_stations(self) -> None:
        org_dect_network = self.webex.org.dect_networks
        if len(org_dect_network) == 0:
            self.skipTest('No DECT Networks')
        self.assertIsInstance(org_dect_network.with_base_stations(), list)
        if len(org_dect_network.with_base_stations()) == 0:
            self.skipTest("No Base Stations")
        network = org_dect_network.with_base_stations()[0]
        self.assertIsInstance(network, wxcadm.dect.DECTNetwork)
        self.assertIsInstance(network.base_stations, list)
        base_station = network.base_stations[0]
        self.assertIsInstance(base_station, wxcadm.dect.DECTBaseStation)
        self.assertIsInstance(base_station.id, str)
        self.assertIsInstance(base_station.mac, str)
        self.assertIsInstance(base_station.number_of_lines_registered, int)

    def test_dect_handsets(self) -> None:
        org_dect_network = self.webex.org.dect_networks
        if len(org_dect_network) == 0:
            self.skipTest('No DECT Networks')
        self.assertIsInstance(org_dect_network.with_base_stations(), list)
        if len(org_dect_network.with_base_stations()) == 0:
            self.skipTest("No Base Stations")
        network = org_dect_network.with_base_stations()[0]
        self.assertIsInstance(network, wxcadm.dect.DECTNetwork)
        self.assertIsInstance(network.handsets, list)
        if len(network.handsets) == 0:
            self.skipTest("No Handsets")
        handset = network.handsets[0]
        self.assertIsInstance(handset, wxcadm.dect.DECTHandset)
        self.assertIsInstance(handset.id, str)
        self.assertEqual(handset.dect_network, network)
        self.assertIsInstance(handset.lines, list)
        line = handset.lines[0]
        self.assertIsInstance(line['memberId'], str)


class TestDectBuildAndRemove(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_full_build(self) -> None:
        with self.subTest("Create DECT Network"):
            new_network = self.webex.org.dect_networks.create(
                'wxcadm Test Network',
                'DBS210',
                location=self.random_location
            )
            self.assertIsInstance(new_network, wxcadm.dect.DECTNetwork)
        with self.subTest("Create Base Station"):
            bs_mac_addr: str = "020000%02x%02x%02x" % (randint(0, 255),
                                                       randint(0, 255),
                                                       randint(0, 255))
            base_stations = new_network.add_base_stations(mac_list=[bs_mac_addr])
            self.assertIsInstance(base_stations, list)
            self.assertIsInstance(base_stations[0], wxcadm.dect.DECTBaseStation)
        with self.subTest("Add Handset"):
            person = choice(self.webex.org.people.webex_calling())
            success = new_network.add_handset('wxcadm', line1=person)
            self.assertTrue(success)
        with self.subTest("Delete Handset"):
            for handset in new_network.handsets:
                new_network.delete_handset(handset)
            self.assertIs(len(new_network.handsets), 0)
        with self.subTest("Delete Base Station"):
            for base in new_network.base_stations:
                new_network.delete_base_station(base)
            self.assertIs(len(new_network.base_stations), 0)
        with self.subTest('Delete DECT Network'):
            success = new_network.delete()
            self.assertTrue(success)



if __name__ == '__main__':
    unittest.main()



from typing import NamedTuple, Optional
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config


class LocationEmergencySettings(NamedTuple):
    """ Enhanced Emergency Call Settings (i.e. RedSky settings) for a Location """
    integration: bool
    """ Whether location data is being sent to RedSky """
    routing: bool
    """ Whether 911 calls are being routed to RedSky"""


@dataclass_json
@dataclass
class OutboundProxy:
    """ Outbound Proxy configuration """
    service_type: str = field(metadata=config(field_name="sipAccessServiceType"))
    """ What the Outbound Proxy is used for """
    dns_type: str = field(metadata=config(field_name="dnsType"))
    """ The type of DNS record represented by the Outbound Proxy """
    proxy_address: str = field(metadata=config(field_name="outboundProxy"))
    """ The hostname of the Outbound Proxy """
    srv_prefix: Optional[str] = field(metadata=config(field_name="srvPrefix"), default=None)
    """ The SRV prefix to use if the DNS type is SRV """
    cname_records: Optional[str] = field(metadata=config(field_name="cnameRecords"), default=None)
    """ Any CNAME records associated with the Outbound Proxy """
    attachment_updated: bool = field(metadata=config(field_name="attachmentUpdated", exclude=lambda t: True), default=False)
    """ Whether the Outbound Proxy was updated """





from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .person import Person
import requests
import base64
import os
from requests_toolbelt import MultipartEncoder
import wxcadm
from wxcadm import log
from .exceptions import *
from .common import *


class CPAPI:
    """The CPAPI class handles API calls using the CP-API, which is the native API used by Webex Control Hub.

    .. note::
            CPAPI instances are normally not instantiated manually and are done automatically with various Org methods.

    .. warning::

            All the methods in this class require an API Access Token with the CP-API scope.
    """

    def __init__(self, org: wxcadm.Org, access_token: str):
        self._parent = org
        self._access_token = access_token
        self._headers = {"Authorization": f"Bearer {access_token}"}

        # Calculate the "customer" ID from the Org ID
        spark_id = decode_spark_id(org.id)
        self._customer = spark_id.split("/")[-1]

        self._url_base = f"https://cpapi-a.wbx2.com/api/v1/customers/{self._customer}/"
        self._server = "https://cpapi-a.wbx2.com"

    def set_global_vm_pin(self, pin: str):
        """Set the Org-wide default VM PIN

        Args:
            pin (str): The PIN to set as the global default

        Returns:
            bool: True if successful

        Raises:
            ValueError: Raised when the PIN value is rejected by Webex, usually because the PIN doesn't comply
                with the security policy.

        """
        log.info("Setting Org-wide default VM PIN")
        payload = {
            "defaultVoicemailPinEnabled": True,
            "defaultVoicemailPin": str(pin)
        }
        r = requests.patch(self._url_base + "features/voicemail/rules",
                           headers=self._headers, json=payload)
        if r.ok:
            return True
        if r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise ValueError

    def clear_global_vm_pin(self):
        log.info("Clearing Org-wide default VM PIN")
        payload = {
            "defaultVoicemailPinEnabled": False,
        }
        r = requests.patch(self._url_base + f"features/voicemail/rules",
                           headers=self._headers, json=payload)
        return r.text

    def reset_vm_pin(self, person: Person, pin: str = None):
        log.info(f"Resetting VM PIN for {person.email}")
        user_id = person.spark_id.split("/")[-1]

        if pin is not None:
            self.set_global_vm_pin(pin)

        r = requests.post(self._url_base + f"users/{user_id}/features/voicemail/actions/resetpin/invoke",
                          headers=self._headers)
        if r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")

        if pin is not None:
            self.clear_global_vm_pin()

        return True

    def get_numbers(self):
        # This method is deprecated and will likely be removed eventually. The Webex API now supports a Numbers GET
        numbers = []
        params = {"limit": 2000, "offset": 0}   # Default values for the numbers pull

        get_more = True     # Bool to let us know to keep pulling more numbers
        next_url = None
        while get_more:
            if next_url is None:
                r = requests.get(self._url_base + f"numbers", headers=self._headers, params=params)
            else:
                r = requests.get(next_url, headers=self._headers)
            if r.status_code == 200:
                response = r.json()
                numbers.extend(response['numbers'])
                if "next" in response['paging']:
                    get_more = True
                    next_url = response['paging']['next']
                else:
                    get_more = False
            elif r.status_code == 403:
                raise TokenError("Your API Access Token doesn't have permission to use this API call")
            else:
                raise APIError(f"The CPAPI numbers call did not return a successful value")

        for number in numbers:
            if "owner" in number:
                if "type" in number['owner'] and number['owner']['type'] == "USER":
                    user_str = f"ciscospark://us/PEOPLE/{number['owner']['id']}"
                    user_bytes = user_str.encode("utf-8")
                    base64_bytes = base64.b64encode(user_bytes)
                    base64_id = base64_bytes.decode('utf-8')
                    base64_id = base64_id.rstrip("=")
                    number['owner']['id'] = base64_id

        return numbers

    def change_workspace_caller_id(self, workspace_id: str, name: str, number: str):
        """ Changes the Caller ID for a Workspace using the CPAPI

        Since Webex Calling uses a different Workspace ID than the CPAPI, this method requires the ID that is recognized
        by CPAPI. The ``name`` and ``number`` arguments take the literal value to be written in the CPAPI payload. The
        methods in other classes that call this method are responsible for sending the correct value to be used.

        Args:
            workspace_id (str): The CPAPI Workspace (Place) ID
            name (str): The value to send to the CPAPI as the "externalCallerIdNamePolicy"
            number (str): The value to send to the CPAPI as the "selected" value

        Returns:
            bool: True on success. False otherwise

        Raises:
            wxcadm.exceptions.APIError: Raised when there is a problem with the API call

        """
        payload = {"externalCallerIdNamePolicy": name,
                   "selected": number}
        r = requests.patch(self._url_base + f"places/{workspace_id}/features/callerid",
                           headers=self._headers, json=payload)
        if r.ok:
            return True
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError("CPAPI failed to update Caller ID for Workspace")

    def get_workspace_calling_location(self, workspace_id: str):
        """ Gets the Location instance associated with a Workspace

        Because Webex Calling uses a different Location than th WorkspaceLocation, sometimes an admin needs to take
        action based on the Calling Location instead. This method takes the ID of the Workspace (from the Device
        instance held in the Webex.Org.devices list) and returns the :class:`Location` instance.

        Args:
            workspace_id (str): The Workspace ID (Device.owner_id) to match on

        Returns:
            Location: The Location instance for the Workspace. None is returned if there is no match.

        Raises:
            wxcadm.exceptions.APIError: Raised when there is a problem getting data from the API

        """
        log.info("CPAPI - Getting Calling Location")
        r = requests.get(self._url_base + f"places/{workspace_id}", headers=self._headers)
        if r.status_code == 200:
            response = r.json()
            location = self._parent.locations.get(name=response['location']['name'])
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError("Something went wrong getting the Workspace Calling Location")
        return location

    def upload_moh_file(self, location_id: str, filename: str):
        log.info("CPAPI - Uploading MOH File")
        # Calculate the "locations" ID from the Org ID
        loc_id_bytes = base64.b64decode(location_id + "===")
        loc_id_decoded = loc_id_bytes.decode("utf-8")
        location_id = loc_id_decoded.split("/")[-1]
        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        must_close = True

        encoder = MultipartEncoder(fields={"file": (upload_as, content, 'audio/wav')})
        log.info(encoder)
        r = requests.post(self._url_base + f"locations/{location_id}/features/musiconhold/actions/announcement",
                          headers={"Content-Type": encoder.content_type, **self._headers},
                          data=encoder)
        if must_close is True:
            content.close()
        if r.ok:
            return True
        elif r.status_code == 403:
            log.debug(f"Request headers: {r.request.headers}")
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError(f"Something went wrong uploading the MOH file: {r.text}")

    def upload_aa_greeting(self, autoattendant, type: str, filename: str):
        log.info("CPAPI - Uploading Auto Attendant Greeting")
        aa_id = autoattendant.spark_id.split("/")[-1]
        location_id = autoattendant.location.spark_id.split("/")[-1]
        if type == "business_hours":
            type = "businessgreetingupload"
        elif type == "after_hours":
            type = "afterhoursgreetingupload"
        else:
            raise ValueError("Valid type values are 'business_hours' and 'after_hours'")

        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        must_close = True
        encoder = MultipartEncoder(fields={"file": (upload_as, content, 'audio/wav')})
        url = self._url_base + f"locations/{location_id}/features/autoattendants/{aa_id}/actions/{type}/invoke"
        r = requests.post(url,
                          params={"customGreetingEnabled": True},
                          headers={"Content-Type": encoder.content_type, **self._headers},
                          data=encoder)
        if must_close is True:
            content.close()
        if r.ok:
            return True
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError(f"Something went wrong uploading the Greeting file: {r.text}")

    def set_custom_aa_greeting(self, autoattendant, type: str, filename: str):
        log.info("CPAPI - Setting Custom AA Greeting")
        aa_id = autoattendant.spark_id.split("/")[-1]
        location_id = autoattendant.location.spark_id.split("/")[-1]
        upload_as = os.path.basename(filename)
        if type == "business_hours":
            type = "businessHoursMenu"
        elif type == "after_hours":
            type = "afterHoursMenu"
        else:
            raise ValueError("Valid type values are 'business_hours' and 'after_hours'")
        config_dict = autoattendant.config[type]
        config_dict['greeting'] = "CUSTOM"
        config_dict['audioFile']['name'] = upload_as
        config_dict['audioFile']['mediaType'] = "WAV"

        payload = {"enabled": autoattendant.config['enabled'], "name": autoattendant.config['name'], **config_dict}

        r = requests.patch(self._url_base + f"locations/{location_id}/features/autoattendants/{aa_id}",
                           headers=self._headers, json=payload)

        if r.ok:
            return True
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError(f"Something went wrong setting the Custom Greeting: {r.text}")

    def set_custom_moh(self, location_id: str, filename: str):
        log.info("CPAPI - Setting Custom MOH")
        # Calculate the "locations" ID from the Org ID
        loc_id_bytes = base64.b64decode(location_id + "===")
        loc_id_decoded = loc_id_bytes.decode("utf-8")
        location_id = loc_id_decoded.split("/")[-1]
        payload = {"callHold": True, "callPark": True, "greeting": "CUSTOM", "description": filename}
        r = requests.patch(self._url_base + f"locations/{location_id}/features/musiconhold",
                           headers=self._headers, json=payload)
        if r.ok:
            return True
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError(f"Something went wrong setting the MOH: {r.text}")

    def set_default_moh(self, location_id: str):
        log.info("CPAPI - Setting Default MOH")
        # Calculate the "locations" ID from the Org ID
        loc_id_bytes = base64.b64decode(location_id + "===")
        loc_id_decoded = loc_id_bytes.decode("utf-8")
        location_id = loc_id_decoded.split("/")[-1]
        # The CPAPI PATCH needs the MOH filename, even for Default, so we need to get what's there already
        r = requests.get(self._url_base + f"locations/{location_id}/features/musiconhold",
                         headers=self._headers)
        current_config = r.json()
        payload = {"callHold": True, "callPark": True, "greeting": "SYSTEM", "description": current_config['description']}
        r = requests.patch(self._url_base + f"locations/{location_id}/features/musiconhold",
                           headers=self._headers, json=payload)
        if r.ok:
            return True
        elif r.status_code == 403:
            raise TokenError("Your API Access Token doesn't have permission to use this API call")
        else:
            raise APIError(f"Something went wrong setting the MOH: {r.text}")



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
from typing import TYPE_CHECKING, Optional

from wxcadm.number import Number, NumberList


class TestNumbers(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())
        # Enable for test debugging
        # wxcadm.console_logging()

    def test_number_lists(self):
        with self.subTest('Org Numbers'):
            org_number_list = self.webex.org.numbers
            self.assertIsInstance(org_number_list, NumberList)
            self.assertIsInstance(org_number_list[0], Number)
            org_number_list = org_number_list.refresh()
            self.assertIsInstance(org_number_list, NumberList)
        with self.subTest('Location Numbers'):
            loc_number_list = self.random_location.numbers
            self.assertIsInstance(loc_number_list, NumberList)
            loc_number_list = loc_number_list.refresh()
            self.assertIsInstance(loc_number_list, NumberList)
        with self.subTest('Get Number by DN'):
            random_number: Optional[Number] = None
            while random_number is None or random_number.phone_number is None:
                random_number = choice(org_number_list)
            self.assertIsInstance(random_number, Number)
            self.assertIsInstance(random_number.phone_number, str)
            got_number = org_number_list.get(phone_number=random_number.phone_number)
            self.assertEqual(random_number, got_number)
        with self.subTest('Get Number by Extension'):
            random_number: Optional[Number] = None
            while random_number is None or random_number.extension is None:
                random_number = choice(org_number_list)
            self.assertIsInstance(random_number, Number)
            got_number = org_number_list.get(extension=random_number.extension)
            self.assertIsInstance(got_number, (Number, list))
            if isinstance(got_number, Number):
                self.assertEqual(random_number, got_number)
        # The following looks for an extension that exists at more than one Location, picks the Location and searches
        # for the extension, confirming it got the right Location
        with self.subTest('Get Number by Extension and Location'):
            seen_extensions = []
            selected_extension = None
            selected_location = None
            for number in org_number_list:
                if number.extension is not None:
                    if number.extension in seen_extensions:
                        selected_extension = number.extension
                        selected_location = number.location
                        break
                    else:
                        seen_extensions.append(number.extension)
            self.assertIsInstance(selected_extension, str)
            self.assertIsInstance(selected_location, wxcadm.Location)
            found = self.webex.org.numbers.get(extension=selected_extension, location=selected_location)
            self.assertIsInstance(found, Number)
            self.assertEqual(selected_extension, found.extension)
            self.assertEqual(selected_location, found.location)
        with self.subTest('Get Number by Owner'):
            random_user: Optional[wxcadm.Person] = None
            org_people = self.webex.org.people.webex_calling()
            while random_user is None or random_user.numbers is None:
                random_user = choice(org_people)
            self.assertIsInstance(random_user, wxcadm.Person)
            number = self.webex.org.numbers.get_by_owner(random_user)
            self.assertIsInstance(number, wxcadm.Number)
            self.assertEqual(random_user, number.owner)





if __name__ == '__main__':
    unittest.main()


import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice, randint


class TestReportsDataPull(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_report_list(self) -> None:
        report_list = self.webex.org.reports
        self.assertIsInstance(report_list, wxcadm.reports.ReportList)
        if len(report_list) > 0:
            self.assertIsInstance(report_list[0], wxcadm.reports.Report)

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import uuid
import json
import srvlookup
import xmltodict
import traceback
import re
import requests
import threading
from typing import Optional, Union
from threading import Thread
import time

import wxcadm
from wxcadm import log
from .common import *
from .exceptions import *
from .xsi_response import XSIResponse


class FeatureAccessCode:
    """ Class for Feature Access Code definitions to make them easier to report on """
    def __init__(self, config: dict):
        """ Init should never be called manually. The instances are created with the :attr:`XSI.fac_list` property """
        self.feature = config['codeName']['$']
        self.code = config['code']['$']
        if 'alternateCode' in config.keys():
            self.alternate_code = config['alternateCode']['$']
        else:
            self.alternate_code = None

class XSIEvents:
    def __init__(self, parent: wxcadm.Org):
        """ Initialize an XSIEvents instance to provide access to XSI-Events

        Args:
            parent (Org): The Webex.Org instance that these Events are associated with. Currently, only Org-level
                events can be monitored. User-level subscriptions will be added later.
        """
        self._parent = parent
        self._headers = parent._headers
        self.events_endpoint = parent.xsi['events_endpoint']
        self.channel_endpoint = parent.xsi['events_channel_endpoint']
        self.xsi_domain = parent.xsi['xsi_domain']
        self.application_id = uuid.uuid4()
        """ The unique Application ID used to identify the application to Webex """
        self.enterprise = self._parent.spark_id.split("/")[-1]
        log.debug(f"Using Enterprise ID: {self.enterprise} to initialize XSIEvents")
        """ The Enterprise ID used to identify the Webex Org to the XSI Events service """
        self.channel_sets = []
        self.queue = None

    def open_channel(self, queue):
        """ Open an XSI Events channel and start pushing Events into the queue.

        For now, this only accepts a Queue instance as the queue argument.

        Args:
            queue (Queue): The Queue instance to place events into

        Returns:
            XSIEventsChannelSet: The :class:`XSIEventsChannelSet` instance that was opened

        """
        self.queue = queue
        log.debug("Initializing Channel Set")
        channel_set = XSIEventsChannelSet(self)
        self.channel_sets.append(channel_set)
        return channel_set


class XSIEventsChannelSet:
    def __init__(self, parent: XSIEvents):
        """ Inititalize an XSIEventsChannelSet instance to manage channels and subscriptions.

        Calling this method, which is normally done with the :meth:`XSIEvents.open_channel()` method, automatically
        creates the channels to each XSP in the region and begins send heartbeat messages on them.

        Args:
            parent (XSIEvents): The XSIEvents instance to which this ChanelSet belongs.
        """
        self.parent = parent
        self._headers = self.parent._headers
        self.id = uuid.uuid4()
        """ The unique ID of the Channel Set """
        log.debug(f"Channel Set ID: {self.id}")
        self.channels = []
        """ List of :py:class:`XSIEventChannel` instances within the Channel Set """
        self.subscriptions = []
        """ List of :py:class:`XSIEventsSubscription` instances associated with the the Channel Set """
        self.queue = self.parent.queue
        """ The Python :py:class:`Queue` that was provided to :py:class`XSIEvents` to queue messages """

        # Now init the channels. Get the SRV records to ensure they get opended with each XSP
        log.debug(f"Getting SRV records for {self.parent.xsi_domain}")
        srv_records = srvlookup.lookup("xsi-client", "TCP", self.parent.xsi_domain)
        self.srv_records = srv_records
        log.debug(f"\t{srv_records}")
        for record in srv_records:
            my_endpoint = re.sub(self.parent.xsi_domain, record.host, self.parent.channel_endpoint)
            my_events_endpoint = re.sub(self.parent.xsi_domain, record.host, self.parent.events_endpoint)
            log.debug(f"Establishing channel with endpoint: {record.host}")
            channel = XSIEventsChannel(self, my_endpoint, my_events_endpoint)
            self.channels.append(channel)

    def restart_failed_channel(self, channel: XSIEventsChannel, wait: int = 0):
        """ Starts a new channel, using an :py:class:`XSIEventsChannel` with an ``active=False`` status as the source
        to determine where the new channel should be established to.

        Args:
            channel (XSIEventsChannel: The :py:class:`XSIEventsChannel` instance that failed.

        Returns:
            bool: True on success, False otherwise.

        """
        if channel.active is True:
            log.warning("Cannot restart an active channel")
            return False
        log.debug(f"[{threading.current_thread().name}] ==== Restarting Failed XSIEventsChannel "
                  f"with endpoint {channel.endpoint.split('/')[2]} in {wait} seconds ====")
        time.sleep(wait)
        new_channel = XSIEventsChannel(self, channel.endpoint, channel.events_endpoint)
        self.channels.append(new_channel)
        # Send a DELETE on the failed channel, just to make sure it is removed from the server
        channel.delete()
        return True

    def audit_channelset(self):
        """ Audit the known channels against what Webex Calling has listed.

        ..note:
            This method is not fully implemented at this time and is being added for future use.

        Returns:

        """
        r = requests.get(self.parent.events_endpoint + f'/v2.0/channelset/{self.id}',
                         headers=self._headers)
        if r.ok:
            log.debug(f"[{threading.current_thread().name}] Channel Audit: {r.text}")

        return True

    @property
    def active_channels(self):
        active_channels = []
        for channel in self.channels:
            if channel.active is True:
                active_channels.append(channel)
        return active_channels

    def subscribe(self, event_package, person: Person = None):
        """ Subscribe to an Event Package over the channel opened with :meth:`XSIEvents.open_channel()`

        Args:
            event_package (str): The name of the Event Package to subscribe to.
            person (Person, optonal): A Person instance to subscribe to the event package for. If not provided,
                the entire Organization will be subscribed.

        Returns:
            XSIEventsSubscription: The Subscription instance. False is returned if the subscription fails.

        """
        log.info(f'Subscribing to {event_package}')
        if len(self.active_channels) == 0:
            log.warning("Cannot subscribe when no channels are active")
            raise XSIError("Cannot subscribe with no active channels")
        subscription = XSIEventsSubscription(self, event_package, person=person)
        if subscription and subscription.active is True:
            self.subscriptions.append(subscription)
            log.debug(f'\tCurrent Subscriptions: {self.subscriptions}')
            return subscription
        else:
            log.warning(f"The subscription did not activate")
            return False

    def unsubscribe(self, subscription_id: str):
        """ Unsubscribe from an Event Package that was previously subscribed to with :meth:`subscribe()`

        Args:
            subscription_id (str): The ID of the XSIEventsSubscription instance

        Returns:
            bool: True if unsubscription was successful, False otherwise

        """
        log.info(f"Unsubscribing subscription: {subscription_id}")
        all_success = True
        for subscription in self.subscriptions:
            if subscription.id == subscription_id or subscription_id.lower() == "all":
                success = subscription.delete()
                if success:
                    self.subscriptions.remove(subscription)
                else:
                    all_success = False
        return all_success

    def close(self):
        """ Close all :py:class:`XSIEventsChannel` and delete all :py:class:`XSIEventsSubscription`

        Returns:
            bool: Always True

        """
        for subscription in self.subscriptions:
            subscription.delete()
        for channel in self.channels:
            log.debug(f"Deleting Channel {channel.id} with {channel.endpoint}")
            channel.delete()
        return True


class XSIEventsChannel:
    def __init__(self, parent: XSIEventsChannelSet, endpoint: str, events_endpoint: str):
        """ The `XSIEventsChannel` handles a single channel in an `XSiEventsChannelSet`

        Upon initialization, the instance will start two threads, one for monitoring the events received across the
        long-running channel and one for sending heartbeats to keep the channel alive.

        .. note::

            This class does not need to be initialized manually. It is done from the :py:class:`XSIEventsChannelSet`
            instance.

        Args:
            parent (XSIEventsChannelSet): The :py:class:`XSIEventsChannelSet` to which this channel belongs
            endpoint (str): The XSI endpoint to connect the long-running channel to
            events_endpoint (str):  The XSI Events endpoint to send heartbeats and acknowledgements to
        """
        self.parent = parent
        self._headers = self.parent.parent._headers
        self.id = ""
        self.endpoint = endpoint
        self.events_endpoint = events_endpoint
        self.queue = self.parent.queue
        self.active = True
        self.needs_restart = False
        self.last_refresh = ""
        self.xsp_ip:str = ''
        self.session = requests.Session()
        self.cookies = None
        self.channel_thread = Thread(target=self._channel_daemon, daemon=True)
        self.heartbeat_thread = Thread(target=self._heartbeat_daemon, daemon=True)

        # Start the channel thread
        self.channel_thread.start()
        # Wait for the Channel to come up, which will mean an .id will be present
        self.wait_count = 0
        while self.id == "":
            self.wait_count += 1
            if self.wait_count >= 20:
                self.active = False
                self.parent.restart_failed_channel(channel=self, wait=1)
            log.debug(f"[{self.wait_count}/20] Waiting for Channel to come up...")
            time.sleep(2)
        self.heartbeat_thread.start()

    def _channel_daemon(self):
        """ A Thread-safe daemon to watch the :py:class:`XSIEventsChannel` for incoming messages.

        .. note::

            This method is called automatically when the `XSIEventsChannel` is initialized and never should be
            called manually

        Returns:
            bool: Will always return True when the channel is closed

        """
        payload = '<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n' \
                  '<Channel xmlns=\"http://schema.broadsoft.com/xsi\">' \
                  f'<channelSetId>{self.parent.id}</channelSetId>' \
                  '<priority>1</priority>' \
                  '<weight>50</weight>' \
                  '<expires>7200</expires>' \
                  f'<applicationId>{self.parent.parent.application_id}</applicationId>' \
                  '</Channel>'

        tid = tracking_id()
        log.debug(f"[{threading.current_thread().name}][{tid}] Sending Channel Start Request (Streaming) "
                  f"to {self.endpoint}/v2.0/channel")
        log.debug(f"\tHeaders: {self._headers}\n\tPayload: {payload}")
        r = self.session.post(
            f"{self.endpoint}/v2.0/channel",
            headers={'TrackingID': tid, **self._headers}, data=payload, stream=True)
        # An issue with the F5 in Oct 2022 showed that we need to handle non-ok responses better
        if not r.ok:
            log.debug(f"\tHeaders: {r.headers}")
            log.debug(f"\tBody: {r.text}")
            log.warning(f"Unable to establish connection with {self.endpoint}")
            ip = 'unknown'
            self.active = False
            self.needs_restart = True
        else:
            # Get the real IP of the remote server so we can send subsequent messages to that
            ip, port = r.raw._connection.sock.getpeername()
        log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                  f"Received response from {ip}. Status code {r.status_code}")
        # log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Headers: {r.headers}")
        log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                  f"Saving Cookies: {r.cookies.get_dict()}")
        self.cookies = r.cookies

        self.xsp_ip = str(ip)
        chars = ""
        self.last_refresh = time.time()
        for char in r.iter_content():
            decoded_char = char.decode('utf-8')
            chars += decoded_char
            # The next line spits character-by-character data out, so it's not really useful outside of development
            # log.debug(chars)
            if "</Channel>" in chars:
                log.debug(f"[{threading.current_thread().name}][{tid}] Channel setup: {chars} from {self.xsp_ip}")
                m = re.search("<channelId>(.+)</channelId>", chars)
                self.id = m.group(1)
                chars = ""
            if "<ChannelHeartBeat xmlns=\"http://schema.broadsoft.com/xsi\"/>" in chars:
                log.debug(f"[{threading.current_thread().name}][{tid}] Heartbeat received on channel {self.id}")
                # Check how long since a channel refresh and refresh if needed
                time_now = time.time()
                if time_now - self.last_refresh >= 3600:
                    log.debug(f"[{threading.current_thread().name}] Refreshing channel: {self.id}")
                    self._refresh_channel()
                    self.last_refresh = time.time()
                # Check any subscriptions and refresh them while we are at it
                for subscription in self.parent.subscriptions:
                    if time_now - subscription.last_refresh >= 3600:
                        log.debug(f"[{threading.current_thread().name}] Refreshing subscription: {subscription.id}")
                        subscription._refresh_subscription()
                        subscription.last_refresh = time.time()
                chars = ""
            if "</xsi:Event>" in chars:
                event = chars
                log.debug(f"[{threading.current_thread().name}][{tid}] Full Event: {event}")
                message_dict = xmltodict.parse(event)
                # Reset ready for new message
                chars = ""
                if message_dict['xsi:Event']['@xsi1:type'] == 'xsi:ChannelTerminatedEvent':
                    # ChannelTerminatedEvent doesn't require or accept an ACK, so we don't need to do that.
                    log.debug(f"[{threading.current_thread().name}][{tid}] "
                              f"Received ChannelTerminatedEvent for channel "
                              f"{message_dict['xsi:Event']['xsi:channelId']} "
                              f"with reason {message_dict['xsi:Event']['xsi:reason']}")
                    # TODO - At some point, we should probably report the channel termination up to see if the channel
                    #        needs to be restarted, or if it happened because we torn it down. The xsi:reason should
                    #        help with that.
                else:
                    self.parent.queue.put(message_dict) # Moved here 2.3.4 to prevent sending TerminatedEvent to Queue
                    try:
                        self._ack_event(message_dict['xsi:Event']['xsi:eventID'])
                    except KeyError:
                        log.debug("xsi:Event received but no xsi:eventID to ACK")
                        log.debug(f"\t{message_dict}")
                        pass
            if decoded_char == "\n":
                # Discard this for now
                # log.debug(f"Discard Line: {chars}")
                chars = ""
        log.debug(f"[{threading.current_thread().name}][{tid}] Channel Loop ended: {self.id}")
        self.active = False
        if self.needs_restart is True:
            self.parent.restart_failed_channel(self, wait=60)
        return True

    def _heartbeat_daemon(self):
        """ A Thread-safe daemon to send heartbeats on the  :py:class:`XSIEventsChannel`

        .. note::

            This method is called automatically when the `XSIEventsChannel` is initialized and never should be
            called manually

        Returns:
            bool: Always returns True when the function completes and the channel is down.

        """
        while self.active:
            tid = tracking_id()
            if self.id is None or self.id == "":
                # If we don't have an ID, we need to kill this and try to restart the channel
                log.warning("No Channel ID found to start heartbeats")
                self.active = False
                self.parent.restart_failed_channel(self)
                continue
            log.debug(f"[{threading.current_thread().name}][{tid}] Sending heartbeat for channel: {self.id}")
            try:
                r = self.session.put(self.events_endpoint + f"/v2.0/channel/{self.id}/heartbeat",
                                headers={'TrackingID': tid, **self._headers}, stream=True)
                ip, port = r.raw._connection.sock.getpeername()
                if r.ok:
                    log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                              f"{ip} - Heartbeat successful")
                    # On success, send a heartbeat every 15 seconds
                    next_heartbeat = 15
                else:
                    log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                              f"{ip} - Heartbeat failed: {r.text} [{r.status_code}]")
                    if r.status_code == 404 or r.status_code == 401:
                        # If the channel can't be found on the server, kill the channel and restart it
                        log.warning(f"Received HTTP {r.status_code}. Marking channel inactive and restarting.")
                        self.active = False
                        self.parent.restart_failed_channel(self)
                        next_heartbeat = 0
                    else:
                        # Losing a heartbeat is ok, but we should try another one sooner than 15 seconds
                        next_heartbeat = 10
            except Exception as e:
                log.debug(f"[{threading.current_thread().name}] Heartbeat failed: {traceback.format_exc()}")
                # If the heartbeat couldn't be sent for some reason, retry sooner than 15 seconds
                next_heartbeat = 10
            #self.parent.audit_channelset()
            time.sleep(next_heartbeat)
        return True

    def _ack_event(self, event_id):
        """ Send an ACK to the XSI Events server to acknowledge the Event was received.

        .. note::

            This method is called automatically by the `channel_daemon()` and should never be called manually

        Args:
            event_id (str): The XSI Event ID that the ACK needs to be sent for.

        Returns:
            bool: True if the ACK was accepted by the server. False otherwise.

        """
        payload = '<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n' \
                  '<EventResponse xmlns=\"http://schema.broadsoft.com/xsi\">' \
                  f'<eventID>{event_id}</eventID>' \
                  '<statusCode>200</statusCode>' \
                  '<reason>OK</reason>' \
                  '</EventResponse>'
        tid = tracking_id()
        log.debug(f"[{threading.current_thread().name}][{tid}] Acking event: {event_id}")
        r = self.session.post(self.events_endpoint + "/v2.0/channel/eventresponse",
                          headers={'TrackingID': tracking_id(), **self._headers}, data=payload)
        if r.ok:
            return True
        else:
            log.debug(f"[{threading.current_thread().name}][{tid}] "
                      f"The ACK was not successful: {r.text}")
            return False

    def _refresh_channel(self):
        """ Send a command to refresh the :py:class:`XSIEventsChannel`

        .. note::

            This method is called automatically by the `heartbeat_daemon()` and should never be called manually

        Returns:
            bool: True on success, False otherwise.

        """
        tid = tracking_id()
        log.debug(f"[{threading.current_thread().name}][{tid}] Refreshing Channel: {self.id}")
        payload = "<Channel xmlns=\"http://schema.broadsoft.com/xsi\"><expires>7200</expires></Channel>"
        r = self.session.put(self.events_endpoint + f"/v2.0/channel/{self.id}",
                         headers={'TrackingID': tracking_id(), **self._headers}, data=payload)
        if r.ok:
            log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                      f"Channel refresh succeeded")
            return True
        else:
            log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                      f"Channel refresh failed: {r.text}")
            return False

    def delete(self):
        """ Delete the :py:class:`XSIEventsChannel` instance

        This method will close the channel and delete it from the server.

        Returns:
            bool: True on success, False otherwise

        """
        tid = tracking_id()
        log.debug(f"[{threading.current_thread().name}][{tid}] Deleting Channel: {self.id}")
        r = self.session.delete(self.events_endpoint + f'/v2.0/channel/{self.id}',
                            headers={'TrackingID': tid, **self._headers})
        log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                  f"Channel Delete received {r.status_code} from server")
        if r.ok:
            self.active = False
            return True
        else:
            log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                      f"Channel Delete response: {r.text}")
            return False


class XSIEventsSubscription:
    def __init__(self, parent: XSIEventsChannelSet, event_package: str, person: Person = None):
        """ Initialize an XSIEventsSubscription

        Initializing the subscription also sends the subscription to the XSI API over the events_endpoint.

        Args:
            parent (XSIEventsChannelSet): The XSIEventsChannelSet instance to issue the subscription for.
            event_package (str): The XSI Event Package to subscribe to.
            person (Person): A Person instance to subscribe to the event package for. If not provided,
                the entire Organization will be subscribed.
        """
        self.id: Optional[str] = None
        self.parent = parent
        if person is None:
            self.target = "serviceprovider"
        else:
            self.target = person
        self.events_endpoint = self.parent.parent.events_endpoint
        self._headers = self.parent._headers
        self.last_refresh = time.time()
        """ The date and time that the subscription was last refreshed """
        self.event_package = event_package
        log.info(f"Subscribing to: {self.event_package} at {self.events_endpoint}")
        payload = '<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n' \
                  '<Subscription xmlns=\"http://schema.broadsoft.com/xsi\">' \
                  f'<event>{self.event_package}</event>' \
                  f'<expires>7200</expires><channelSetId>{self.parent.id}</channelSetId>' \
                  f'<applicationId>{self.parent.parent.application_id}</applicationId></Subscription>'
        if person is None:
            r = requests.post(self.events_endpoint + f"/v2.0/serviceprovider/{self.parent.parent.enterprise}",
                              headers={"TrackingId": tracking_id(), **self._headers}, data=payload)
            # Adding the following debug to help with failed subscriptions.
            log.debug("Subscribe API Call:")
            log.debug(f"\tURL: {self.events_endpoint}/v2.0/serviceprovider/{self.parent.parent.enterprise}")
            log.debug(f"\tHeaders: {self._headers}")
            log.debug(f"\tPayload: {payload}")
        else:
            if isinstance(person, wxcadm.person.Person):
                # Make sure we have the Person's XSI Profile
                user_id = person.spark_id.split("/")[-1]
                r = requests.post(self.events_endpoint + f"/v2.0/user/{user_id}",
                                  headers=self._headers, data=payload)
            else:
                raise ValueError("The user argument requires a Person instance")
        if r.ok:
            response_dict = xmltodict.parse(r.text)
            log.debug(f"Response: {response_dict}")
            self.id = response_dict['Subscription']['subscriptionId']
            log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Subscription ID: {self.id}")
            self.active = True
        else:
            log.warning(f"Subscription failed: [{r.status_code}] {r.text}")
            self.active = False

    def _refresh_subscription(self):
        """ Refresh the subscription

        .. note::

            This method is called automatically by the channels based on the subscription expiration

        Returns:
            bool: True on success, False otherwise

        """

        payload = "<Subscription xmlns=\"http://schema.broadsoft.com/xsi\"><expires>7200</expires></Subscription>"
        r = requests.put(self.events_endpoint + f"/v2.0/subscription/{self.id}",
                         headers={'TrackingID': tracking_id(), **self._headers}, data=payload)
        if r.ok:
            log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                      f"Subscription refresh succeeded")
            return True
        else:
            log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                      f"Subscription refresh failed: {r.text}")
            return False

    def delete(self):
        """ Delete the subscription from the server

        Once the subscription is deleted, no events for that subscription will be received.

        Returns:
            bool: True on success, False otherwise

        """
        log.debug(f"Deleting XSIEventsSubscription: {self.id}, {self.event_package}")
        r = requests.delete(self.events_endpoint + f"/v2.0/subscription/{self.id}",
                            headers={'TrackingID': tracking_id(), **self._headers})
        if r.ok:
            log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Subscription delete succeeded")
            return True
        else:
            log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Subscription delete failed: {r.text}")
            return False


class XSICallQueue:
    def __init__(self, call_queue_id: str, org: wxcadm.Org):
        """ Initialize an instance to control calls within a Webex Calling Call Queue.

        This class is used primarily when monitoring XSI Events where there is a need to control calls that come
        into a Call Queue. Since Call Queues are "virtual" users within the call control, there are special methods
        required to deal with those calls.

        Args:
            call_queue_id (str): The ID of the Call Queue as shown in the ``xsi:targetId`` of an XSI Event
            org (Org): The instance of the wxcadm.Org class (i.e. "webex.org"). This is needed to provide the right
                authentication and URLs for the XSI service.
        """
        log.info(f"Starting XSICallCenter for {call_queue_id}")
        self.id = call_queue_id
        self.org = org
        self._calls = []
        self._headers = org._headers
        self._params = org._params

    def attach_call(self, call_id: str):
        """ Attach an active call to the XSICallQueue instance

        Attaching a call provides a :class:`wxcadm.Call` instance that can be used to control the call. Note that not
        all call controls are available to a Call Queue call, but the basic ones, including Transfer, are available.

        Args:
            call_id (str): The callId from the XSI Events message

        Returns:
            Call: The :class:`wxcadm.Call` instance that was attached

        """
        call = Call(self, id=call_id)
        self._calls.append(call)
        return call


class XSI:
    def __init__(self, parent, get_profile: bool = False, cache: bool = False):
        """The XSI class holds all of the relevant XSI data for a Person

        Args:
            parent (Person): The Person who this XSI instance belongs to
            get_profile (bool): Whether or not to automatically get the XSI Profile
            cache (bool): Whether to cache the XSI data (True) or pull it "live" every time (**False**)

        """
        log.info(f"Initializing XSI instance for {parent.email}")
        # First we need to get the XSI User ID for the Webex person we are working with
        log.info("Getting XSI identifiers")
        user_spark_id = decode_spark_id(parent.id)
        self.id = user_spark_id.split("/")[-1]

        # Inherited attributes
        self.xsi_endpoints = parent._parent.xsi
        self._cache = cache

        # API attributes
        self._headers = {"Content-Type": "application/json",
                         "Accept": "application/json",
                         "X-BroadWorks-Protocol-Version": "25.0",
                         **parent._headers}
        self._params = {"format": "json"}

        # Attribute definitions
        self._calls: list = []
        self._profile: dict = {}
        """The XSI Profile for this Person"""
        self._registrations: dict = {}
        """The Registrations associated with this Person"""
        self.services = {}
        self._alternate_numbers: dict = {}
        """The Alternate Numbers for the Person"""
        self._anonymous_call_rejection: dict = {}
        """The Anonymous Call Rejection settings for this Person"""
        self._executive_assistant: dict = {}
        """The Executive Assistant settings for this Person"""
        self._executive: dict = {}
        """The Executive settings for this Person"""
        self._single_number_reach: dict = {}
        """The SNR (Office Anywhere) settings for this Person"""
        self._monitoring: dict = {}
        """The BLF/Monitoring settings for this Person"""
        self.conference: object = None

        # Get the profile if we have been asked to
        if get_profile:
            self._profile = self.profile

    def new_call(self, address: str = None, phone: str = 'All', aor: Optional[str] = None):
        """Create a new Call instance

        Args:
            address (str, optional): The address to originate a call to
            phone (str, optional): The phone to use for the call. Valid values are ``All``, ``Primary``,
                or ``SharedCallAppearance``. The SharedCallAppearance argument also requires an address-or-record
                ``aor`` argument with the AoR of the Shared device.
            aor (str, optional): The address-of-record of the Shared device to place the call from

        Returns:
            Call: The Call instance

        """
        # If we got an address, pass it to the new instance
        call = Call(self)
        if address is not None:
            call.originate(address=address, phone=phone, aor=aor)
        self._calls.append(call)
        return call

    def new_conference(self, calls: list, comment: str = ""):
        """
        Crates a new Conference instance. A user can only have one Conference instance, so this will replace any
        previous Conference. At the moment, this **should not be called directly** and will be done dynamically by
        a Call.conference()

        Args:
            calls (list): A list of Call IDs involved in this conference. A conference is always started with only
                two Call IDs. Call IDs after the first two will be ignored.
            comment (str, optional): An optional text comment for the conference

        Returns:
            The instance of the Conference class

        """
        self.conference = Conference(self, calls, comment)
        return self.conference

    def answer_call(self, call_id: str = None):
        """ Answer an incoming call

        If the call_id is not provided, the method will try and answer the latest/only call for the user.

        Args:
            call_id (str, optional): The call_id to answer

        Returns:
            Call: The :class:`Call` instance of the answered call. None is returned if no calls can be found.

        """
        call = self.calls[-1]
        if call is not None:
            call.resume()
            return call
        else:
            return None

    @property
    def calls(self):
        """
        Get the list of active calls and creates Call instances. Also destroys any Call instances that are no longer
        valid.

        Returns:
            list[Call]: List of Call instances

        """
        # First wipe out all of the existing instances
        for call in self._calls:
            del call
        self._calls.clear()
        calls_data: dict = self.__get_xsi_data(f"/v2.0/user/{self.id}/calls")
        log.debug(f"Calls Data: {calls_data}")
        if "call" not in calls_data['Calls']:
            self._calls = []
            return self._calls
        if type(calls_data['Calls']['call']) is dict:
            this_call = Call(self, id=calls_data['Calls']['call']['callId']['$'])
            self._calls.append(this_call)
        elif type(calls_data['Calls']['call']) is list:
            for call in calls_data['Calls']['call']:
                this_call = Call(self, id=call['callId']['$'])
                self._calls.append(this_call)
        return self._calls

    def attach_call(self, call_id: str):
        """ Attach an active call to the XSICallQueue instance

        Attaching a call provides a :class:`wxcadm.Call` instance that can be used to control the call. Note that not
        all call controls are available to a Call Queue call, but the basic ones, including Transfer, are available.

        Args:
            call_id (str): The callId from the XSI Events message

        Returns:
            Call: The :class:`wxcadm.Call` instance that was attached

        """
        call = Call(self, id = call_id)
        self._calls.append(call)
        return call

    def __get_xsi_data(self, url, params: Optional[dict] = None):
        if params is not None:
            params = {**params, **self._params}
        log.debug(f"Sending API Call: {self.xsi_endpoints['actions_endpoint'] + url}")
        log.debug(f"\tHeaders: {self._headers}")
        log.debug(f"\tParams: {params}")
        r = requests.get(self.xsi_endpoints['actions_endpoint'] + url, headers=self._headers, params=params)
        log.debug(f"Response: {r.status_code}")
        if r.status_code == 200:
            try:
                response = r.json()
                # As of the addition of directory() in 3.1.0, we have to be prepared for paginated responses.
                # There is a chance to add that here so that the invoking method doesn't have to care about it, but
                # for now we are going to just return the dict we get and let the other method decide if it needs more
                # records and send this method the right params
            except json.decoder.JSONDecodeError:
                response = r.text
            return_data = response
        elif r.status_code == 404:
            return_data = False
        else:
            log.warning(f'Headers: {r.headers}')
            log.warning(f'Body: {r.text}')
            return_data = False
        r.close()
        return return_data

    @property
    def executive(self):
        """The Executive Assistant settings for this Person"""
        if not self._executive or not self._cache:
            self._executive = self.__get_xsi_data(f"/v2.0/user/{self.id}/services/Executive")
        return self._executive

    @property
    def executive_assistant(self):
        """The Executive Assistant settings for this Person"""
        if not self._executive_assistant or not self._cache:
            self._executive_assistant = self.__get_xsi_data(f"/v2.0/user/{self.id}/services/ExecutiveAssistant")
        return self._executive_assistant

    @property
    def monitoring(self):
        """The Monitoring/BLF settings for this person"""
        if not self._monitoring or not self._cache:
            self._monitoring = self.__get_xsi_data(f"/v2.0/user/{self.id}/services/BusyLampField")
        return self._monitoring

    @property
    def single_number_reach(self):
        """The SNR (Office Anywhere) settings for this Person"""
        if not self._single_number_reach or not self._cache:
            self._single_number_reach = \
                self.__get_xsi_data(f"/v2.0/user/{self.id}/services/BroadWorksAnywhere")
        return self._single_number_reach

    @property
    def anonymous_call_rejection(self):
        """The Anonymous Call Rejection settings for this Person"""
        if not self._anonymous_call_rejection or not self._cache:
            self._anonymous_call_rejection = \
                self.__get_xsi_data(f"/v2.0/user/{self.id}/services/AnonymousCallRejection")
        return self._anonymous_call_rejection

    @property
    def alternate_numbers(self):
        """The Alternate Numbers for this Person"""
        if not self._alternate_numbers or not self._cache:
            self._alternate_numbers = \
                self.__get_xsi_data(f"/v2.0/user/{self.id}/services/AlternateNumbers")
        return self._alternate_numbers

    @property
    def profile(self):
        """The XSI Profile for this Person"""
        log.debug("Getting XSI Profile")
        if not self._profile or not self._cache:
            profile_data: dict = \
                self.__get_xsi_data(f"/v2.0/user/{self.id}/profile")
            # The following is a mapping of the raw XSI format to the profile attribute
            log.debug(f"User Profile: {profile_data}")
            if profile_data:
                self._profile['registrations_url'] = profile_data['Profile']['registrations']['$']
                self._profile['schedule_url'] = profile_data['Profile']['scheduleList']['$']
                self._profile['fac_url'] = profile_data['Profile']['fac']['$']
                self._profile['country_code'] = profile_data['Profile']['countryCode']['$']
                self._profile['user_id'] = profile_data['Profile']['details']['userId']['$']
                self._profile['group_id'] = profile_data['Profile']['details']['groupId']['$']
                self._profile['service_provider'] = profile_data['Profile']['details']['serviceProvider']['$']
                # Not everyone has a number and/or extension, so we need to check to see if there are there
                if "number" in profile_data['Profile']['details']:
                    self._profile['number'] = profile_data['Profile']['details']['number']['$']
                if "extension" in profile_data['Profile']['details']:
                    self._profile['extension'] = profile_data['Profile']['details']['extension']['$']
                self._profile['raw'] = profile_data
            else:
                self._profile = None
        return self._profile

    @property
    def registrations(self):
        """The device registrations associated with this Person"""
        if not self._registrations or not self._cache:
            # If we don't have a registrations URL, because we don't have the profile, go get it
            if "registrations_url" not in self._profile:
                self._profile = self.profile
            self._registrations = self.__get_xsi_data(self._profile['registrations_url'])
        return self._registrations

    @property
    def fac_list(self):
        # If we don't have a FAC URL, go get it
        if "fac_url" not in self._profile:
            self.profile
        r = requests.get(self.xsi_endpoints['actions_endpoint'] + self._profile['fac_url'],
                         headers=self._headers, params=self._params)
        response = r.json()
        fac_list = []
        for entry in response['FAC']['featureAccessCode']:
            fac_list.append(FeatureAccessCode(entry))
        return fac_list

    def get_services(self):
        # TODO There are still some services that we should collect more data for. For example, BroadWorks
        #       Anywhere has Locations that aren't pulled without a separate call.

        r = requests.get(self.xsi_endpoints['actions_endpoint'] + "/v2.0/user/" + self.id + "/services",
                         headers=self._headers, params=self._params)
        response = r.json()
        self.services['list'] = response['Services']['service']
        # Now that we have all of the services, pulling the data is pretty easy since the URL
        # is present in the response. Loop through the services and collect the data
        # Some services have no config so there is no URI and we'll just populate them as True
        for service in self.services['list']:
            if "uri" in service:
                r = requests.get(self.xsi_endpoints['actions_endpoint'] + service['uri']['$'],
                                 headers=self._headers, params=self._params)
                # Getting well-formatted JSON doesn't always work. If we can decode the JSON, use it
                # If not, just store the raw text. At some point, it would make sense to parse the text
                # and build the dict directly
                try:
                    response = r.json()
                except json.decoder.JSONDecodeError:
                    response = r.text
                self.services[service['name']['$']] = response
            else:
                self.services[service['name']['$']] = True
        return self.services

    def directory(self,
                  type: Optional[str] = 'Enterprise',
                  first_name: Optional[str] = None,
                  last_name: Optional[str] = None,
                  name: Optional[str] = None,
                  user_id: Optional[str] = None,
                  group_id: Optional[str] = None,
                  number: Optional[str] = None,
                  extension: Optional[str] = None,
                  mobile_number: Optional[str] = None,
                  department: Optional[str] = None,
                  email: Optional[str] = None,
                  any_match: Optional[bool] = False,) -> list[dict]:
        """ Search the Webex Calling directories

        When search filters are applied as arguments, the directory will be searched by those values. If the desire is
        to match any one of the values (logical OR), the ``any_match`` parameter should also be set to True. Note that
        each directory type supports its own filter criteria and not all criteria are available across all directories.
        The following table shows the filter arguments that are available for each type.

        Note that all directory filter arguments are case-sensitive within Webex Calling. If you want to perform a
        case-insensitive search, append "/i" to the search string. For example, ``first_name='John/i'`` will match
        "John", "john" and "JOHN".

        .. list-table:: Directory Types
            :widths: 25 75
            :header-rows: 1

            * - Directory Type
              - Available Filters
            * - Enterprise
              - first_name, last_name, name, user_id, group_id, number, extension, mobile_number, department, email
            * - Group
              - first_name, last_name, name, user_id, group_id, number, extension, mobile_number, department, email
            * - Personal
              - name, number


        Args:
            type (str, Optional): The type of diectory to search. Valid values are ``Enterprise``,
                ``Group``, and ``Personal``. Defaults to ``Enterprise``.
            first_name (str, optional): The First Name field in the directory
            last_name (str, optional): The Last Name field in the directory
            name (str, optional): The "combined" name field that allows a search based on First Name and Last Name
            user_id (str, optional): The User ID used on the Webex Calling call control platform. This is the User ID
                that is displayed for the user across all of the XSI APIs
            group_id (str, optional): The Group ID used on the Webex Calling call control platform. This ID is unique
                to the Webex Calling Location and can be used to find all users at a specific Location.
            number (str, optional): The Webex Calling phone number
            extension (str, optional): The Webex Calling extension
            mobile_number (str, optional): The user's mobile number, if populated by Directory Sync
            department (str, optional): The user's department, if populated by Directory Sync
            email (str, optional): The user's email address
            any_match (bool, optional): When True, all arguments will be treated uniquely (OR) and results will be
                returned if any argument matches. For exmaple, if the arguments
                ``first_name=Joe,last_name=Smith,any_match=True`` are passed, the results will include Joe Smith, but
                will also include Joe James and Lisa Smith. Defaults to ``False``.

        Returns:
            list[dict]: A list of matching records, represented as a dictionary. If there are no matches, an empty
                list is returned. If an invalid directory type is requested, None is returned.

        """
        type = type.title()
        log.info(f"Getting {type} directory")
        log.debug(f"\tArgs: {locals()}")
        if type.upper() in ['ENTERPRISE', 'GROUP']:
            params = {
                'firstName': first_name,
                'lastName': last_name,
                'name': name,
                'userId': user_id,
                'groupId': group_id,
                'number': number,
                'extension': extension,
                'mobileNo': mobile_number,
                'department': department,
                'emailAddress': email,
                'searchCriteriaModeOr': any_match,
            }
        elif type.upper() in ['PERSONAL']:
            params = {
                'name': name,
                'number': number
            }
        else:
            log.warning(f"Directory type {type} is not valid")
            return None

        # Set some vars to keep track of the number of records we get and how many more we expect
        more_records = True
        num_records = 0
        next_index = 1
        get_count = 50
        return_records = []
        while more_records is True:
            response = self.__get_xsi_data(f'/v2.0/user/{self.id}/directories/{type}',
                                           params={'start': next_index, 'results': get_count, **params})
            num_records += int(response[type]['numberOfRecords']['$'])
            total_records = int(response[type]['totalAvailableRecords']['$'])
            log.debug(f"Received {num_records}/{total_records} records")
            if int(response[type]['numberOfRecords']['$']) == 0: # No records returned
                return return_records
            if type.upper() in ['ENTERPRISE', 'GROUP']:
                key = f"{type.lower()}Directory"
                if isinstance(response[type][key]['directoryDetails'], dict):   # If we get a dict instead of a list
                    return_records.append(response[type][key]['directoryDetails'])
                else:
                    for entry in response[type][key]['directoryDetails']:
                        return_records.append(entry)
            if type.upper() in ['PERSONAL']:
                if isinstance(response[type]['entry'], dict):   # Single entry returned
                    return_records.append(response[type]['entry'])
                else:
                    for entry in response[type]['entry']:
                        return_records.append(entry)

            # Check and see if we need to get more records
            if num_records < total_records:
                more_records = True
                next_index += int(response[type]['numberOfRecords']['$'])
            else:
                more_records = False

        return return_records


class Call:
    """
    The Call class represents a call for a person. Since Webex supports calls in the Webex API as well as XSI API,
    the class supports both styles. When initialized, the parent instance is checked to see if it is a Person
    instance or an XSI instance. At the moment, the Webex API only supports user-scoped call control, so most of the
    development focus right now is the XSI API, which is more feature-rich

    """

    def __init__(self, parent: Union[Person, XSI, XSICallQueue],
                 id: str = "", address: str = "", user_id: str = ""):
        """Inititalize a Call instance for a Person

        Args:
            parent (XSI): The Person or XSI instance that owns this Call
            id (str, optional): The Call ID of a known call. Usually only done during a XSI.calls method
            address (str, optional): The address to originate a call to when the instance is created

        Returns:
            Call: This Call instance

        """
        self._parent = parent
        """The Person or XSI instance that owns this Call"""
        if user_id:
            self._userid = user_id
        else:
            self._userid: str = self._parent.id
        """The Person or XSI ID inherited from the parent"""
        self._headers = self._parent._headers
        self._params = self._parent._params
        self._url: str = ""
        self.id: str = id
        """The Call ID for this call"""
        self._external_tracking_id: str = ""
        """The externalTrackingId used by XSI"""
        self._status: dict = {}
        """The status of the call"""
        self.held: bool = False
        """ Whether or not the call is on hold """
        self._transfer_call = None
        if type(self._parent) is wxcadm.person.Person:
            # This is where we set things based on whether the parent is a Person
            self._url = _url_base
            self.type = "Person"
            pass
        elif type(self._parent) is wxcadm.xsi.XSICallQueue:
            self._url = self._parent.org.xsi['actions_endpoint'] + f"/v2.0/callcenter/{self._userid}/calls"
            self.type = "CallQueue"
        elif type(self._parent) is wxcadm.xsi.XSI:
            # The Call parent is XSI
            self._url = self._parent.xsi_endpoints['actions_endpoint'] + f"/v2.0/user/{self._userid}/calls"
            self.type = "XSI"
        elif type(self._parent) is wxcadm.xsi.Call:
            # Another Call created this Call instance (probably for a transfer or conference
            self._url = self._parent.xsi_endpoints['actions_endpoint'] + f"/v2.0/user/{self._parent._userid}/calls"
            self.type = "Call"
        elif type(self._parent) is wxcadm.org.Org:
            # Basically manually creating a call instance, probably based on an XSI Event or some other way
            # that we determined the call ID and the user it was for
            self._url = self._parent.xsi['actions_endpoint'] + f"/v2.0/user/{self._userid}/calls"
            self.type = "Org/Other"

        if address:
            self.originate(address)

    def originate(self, address: str,
                  comment: str = "",
                  phone: str = 'All',
                  aor: Optional[str] = None,
                  executive: str = None):
        """Originate a call on behalf of the Person

        The optional ``executive`` argument takes an extension or phone number and allows the call to be placed
        on behalf of an executive to which the Person is assigned as an executive assistant. If the Exectuve call is
        not allowed by the system, an :exc:`NotAllowed` is raised.

        Args:
            address (str): The address (usually a phone number) to originate the call to
            comment (str, optional): Text comment to attach to the call
            phone (str, optional): The phone to use for the call. Valid values are ``All``, ``Primary``,
                or ``SharedCallAppearance``. The SharedCallAppearance argument also requires an address-or-record
                ``aor`` argument with the AoR of the Shared device.
            aor (str, optional): The address-of-record of the Shared device to place the call from
            executive (str, optional): The phone number or extension of the Executive to place the call on behalf of

        Returns:
            bool: True when the call was successful

        Raises:
            wxcadm.exceptions.NotAllowed: Raised when the Person is not able to place the call for an Executive

        """
        if executive is not None:
            log.info(f"Originating a call to {address} for {self._userid} on behalf of Exec {executive}")
            # TODO: The API call will fail if the Assistant can't place the call for the Exec, but we should
            #   really check that first and not waste the API call (although those take API calls, too)
            params = {"address": address, "executiveAddress": executive}
            r = requests.post(self._url + "/ExecutiveAssistantInitiateCall", headers=self._headers, params=params)
            if r.status_code == 201:
                response = r.json()
                self.id = response['CallStartInfo']['callId']['$']
                self._external_tracking_id = response['CallStartInfo']['externalTrackingId']['$']
                return True
            else:
                raise NotAllowed("Person is not allowed to place calls on behalf of this executive")
        else:
            log.info(f"Originating a call to {address} for {self._userid}")
            params = {"address": address, "info": comment, "location": phone}
            if aor is not None:
                params['locationAddress'] = aor
            r = requests.post(self._url + "/new", headers=self._headers, params=params)
            if r.status_code == 201:
                response = r.json()
                self.id = response['CallStartInfo']['callId']['$']
                self._external_tracking_id = response['CallStartInfo']['externalTrackingId']['$']
                return True
            else:
                return False

    def exec_push(self):
        """Pushes the active Executive Assistant call to the Executive

        This method will only complete if the following conditions are met:
        * The user is an Assistant
        * The call must be active and answered

        Returns:
            bool: True if the push was successful

        Raises:
            wxcadm.exceptions.NotAllowed: Raised when the call does not meet the conditions to be pushed

        """
        r = requests.put(self._url + f"/{self.id}/ExecutiveAssistantCallPush", headers=self._headers)
        return XSIResponse(r)

    def hangup(self, decline: bool = False):
        """Hang up the call

        Returns:
            bool: Whether the command was successful

        """
        if decline is True:
            params = {"decline": decline}
        else:
            params = {}
        log.info(f"Hanging up call ID: {self.id}")
        r = requests.delete(self._url + f"/{self.id}",
                            headers=self._headers, params=params)
        return XSIResponse(r)

    @property
    def status(self):
        """The status of the call

        Returns:
            dict::

                {
                'network_call_id' (str): The unique identifier for the Network side of the call
                'personality'(str): The user personalty (Originator or Terminator)
                'state' (str): The state of the call
                'remote_party' (dict): {
                    'address' (str): The address of the remote party
                    'call_type' (str): The call type
                    }
                'endpoint' (dict): {
                    'type' (str): The type of endpoint being used
                    'AoR' (str): The Address of Record for the endpoint
                    }
                'appearance' (str): The Call Appearance number
                'start_time' (str): The UNIX timestanp of the start of the call
                'answer_time' (str): The UNIX timestamp when the call was answered
                'status_time' (str): The UNIX timestamp of the status response
                }

        """
        log.info(f"Getting call status")
        r = requests.get(self._url + f"/{self.id}",
                         headers=self._headers)
        response = r.json()
        log.debug(f"Call Status response: {response}")
        if r.status_code == 200:
            return_data = {}
            ## Added key checking 3.0.4
            call_status = response['Call']
            if 'networkCallId' in call_status:
                return_data['network_call_id'] = call_status['networkCallId']['$']
            if 'personality' in call_status:
                return_data['personality'] = call_status['personality']['$']
            if 'state' in call_status:
                return_data['state'] = call_status['state']['$']
            if 'remoteParty' in call_status:
                return_data['remote_party'] = {
                    "address": call_status['remoteParty']['address']['$'],
                    "call_type": call_status['remoteParty']['callType']['$']
                }
            if 'endpoint' in call_status:
                return_data['endpoint'] = {
                    "type": call_status['endpoint']['@xsi1:type'],
                    "aor": call_status['endpoint']['addressOfRecord']['$']
                }
            if 'appearance' in call_status:
                return_data['appearance'] = call_status['appearance']['$']
            if 'diversionInhibited' in call_status:
                return_data['diversion_inhibited'] = call_status['diversionInhibited']
            if 'startTime' in call_status:
                return_data['start_time'] = call_status['startTime']['$']
            if 'answerTime' in call_status:
                return_data['answer_time'] = call_status['answerTime']['$']
            else:
                return_data['answer_time'] = None
            return_data['status_time'] = int(time.time())
            return return_data
        else:
            return False

    def transfer(self, address: str, type: str = "blind"):
        """Transfer the call to the selected address.

        Type of transfer can be controlled with `type` param. VM transfers will transfer the call directly to the voice
        mail of the address, even if the address is the user's own address. Attended transfers require a subsequent call
        to `finish_transfer()` when the actual transfer should happen.

        Args:
            address (str): The address (usually a phone number or extension) to transfer the call to
            type (str): ['blind','vm','attended']:
                The type of transfer.

        Returns:
            bool: True if successful. False if unsuccessful

        """
        log.info(f"Transferring call {self.id} to {address} for {self._userid}")
        # Set the address param to be passed to XSI
        if self.type == "CallQueue":
            params = {"phoneno": address}
        else:
            params = {"address": address}
        # Handle an attended transfer first. Anything else is assumed to be blind
        if type.lower() == "attended":
            # Attended transfer requires the first call to be put on hold and the second call to be
            # placed, so those are here. A separate call to finish_transfer will be required when the transfer should
            # happen.
            self.hold()
            self._transfer_call = self._parent.new_call()
            self._transfer_call.originate(address)
            return True
        elif type.lower() == "vm":
            r = requests.put(self._url + f"/{self.id}/VmTransfer", headers=self._headers, params=params)
            return XSIResponse(r)
        elif type.lower() == "mute":
            r = requests.put(self._url + f"/{self.id}/MuteTransfer", headers=self._headers, params=params)
            return XSIResponse(r)
        else:
            r = requests.put(self._url + f"/{self.id}/BlindTransfer", headers=self._headers, params=params)
            return XSIResponse(r)

    def finish_transfer(self):
        """
        Complete an Attended Transfer. This method will only complete if a `transfer(address, type="attended")`
        has been done first.

        Returns:
            bool: Whether the transfer completes

        """
        log.info("Completing transfer...")
        r = requests.put(self._url + f"/{self.id}/ConsultTransfer/{self._transfer_call.id}", headers=self._headers)
        return XSIResponse(r)

    def conference(self, address: str = ""):
        """
        Starts a multi-party conference. If the call is already held and an attended transfer is in progress,
        meaning the user is already talking to the transfer-to user, this method will bridge the calls.

        When an address is passed, the existing call will be placed on hold, the second call will be originated, and the
        conference will be connected as soon as the second party answers. If the desired behavior is not to connect the
        parties until later, the calls should be created separately and bridged later.

        Args:
            address (str, optional): The address (usually a phone number or extension) to conference to. Not needed
                when the call is already part of an Attended Transfer

        Returns:
            bool: True if the conference is successful

        """
        # First, check to see if the call is already part of an attended transfer. If so, just build the conference
        # based on the two call IDs
        if self._transfer_call:
            xml = f"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" \
                  f"<Conference xmlns=\"http://schema.broadsoft.com/xsi\">" \
                  f"<conferenceParticipantList>" \
                  f"<conferenceParticipant>" \
                  f"<callId>{self.id}</callId>" \
                  f"</conferenceParticipant>" \
                  f"<conferenceParticipant>" \
                  f"<callId>{self._transfer_call.id}</callId>" \
                  f"</conferenceParticipant>" \
                  f"</conferenceParticipantList>" \
                  f"</Conference>"
            # Building the XML by hand for right now. Probably going to replace it with something JSON-friendly
            headers = self._headers
            headers['Content-Type'] = "application/xml; charset=UTF-8"
            r = requests.post(self._url + f"/Conference", headers=headers, data=xml)
            if r.status_code in [200, 201, 204]:
                return self._parent.new_conference([self.id, self._transfer_call.id])
            else:
                return False
        else:
            # Put the current call on hold if it isn't already
            if self.held is False:
                self.hold()
            # Build the second call
            second_call = self._parent.new_call(address=address)
            # Get the status of the second call. We cannot complete the conference until it is answered.
            call_state = second_call.status.get('state', 'unknown')
            while call_state != 'Active':
                call_state = second_call.status.get('state', 'unknown')

            xml = f"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" \
                  f"<Conference xmlns=\"http://schema.broadsoft.com/xsi\">" \
                  f"<conferenceParticipantList>" \
                  f"<conferenceParticipant>" \
                  f"<callId>{self.id}</callId>" \
                  f"</conferenceParticipant>" \
                  f"<conferenceParticipant>" \
                  f"<callId>{second_call.id}</callId>" \
                  f"</conferenceParticipant>" \
                  f"</conferenceParticipantList>" \
                  f"</Conference>"
            headers = self._headers
            headers['Content-Type'] = "application/xml; charset=UTF-8"
            r = requests.post(self._url + f"/Conference", headers=headers, data=xml)
            if r.status_code in [200, 201, 204]:
                return self._parent.new_conference([self.id, second_call.id])
            else:
                return False

    def send_dtmf(self, dtmf: str):
        """Transmit DTMF tones outbound

        Args:
            dtmf (str): The string of dtmf digits to send. Accepted digits 0-9, star, pound. A comma will pause
                between digits (i.e. "23456#,123")

        Returns:
            bool: True if the dtmf was sent successfully
        """
        params = {"playdtmf": str(dtmf)}
        r = requests.put(self._url + f"/{self.id}/TransmitDTMF", headers=self._headers, params=params)
        return XSIResponse(r)

    def hold(self):
        """Place the call on hold

        Returns:
            bool: Whether the hold command was successful

        """
        r = requests.put(self._url + f"/{self.id}/Hold", headers=self._headers)
        self.held = True
        return XSIResponse(r)

    def resume(self):
        """Resume a call that was placed on hold

        Returns:
            bool: Whether the command was successful

        """
        r = requests.put(self._url + f"/{self.id}/Talk", headers=self._headers)
        self.held = False
        return XSIResponse(r)

    def answer(self):
        r = requests.put(self._url + f"/{self.id}/Talk", headers=self._headers)
        self.held = False
        return XSIResponse(r)

    def park(self, extension: str = None):
        """Park the call

        When called with the ``extension`` argument, the call will be parked using the Call Park Extension feature at
        the chosen extension. When called without ``extension``, the call will be parked with the Group Call Park
        feature, which assigns an extension automatically. Note that Group Call Park requires the Person to be part
        of the Park Group. If a Group Call Park is attemped for a user that isn't park of a Park Group, a
        :exc:`NotAllowed` exception will be raised.

        Args:
            extension (str, optional): The extension to park the call against

        Returns:
            str: The extension that the call is parked against

        Raises:
            wxcadm.exceptions.NotAllowed: Raised when the user is not part of a Park Group or the extension is already busy

        """
        if extension is None:
            r = requests.put(self._url + f"/{self.id}/GroupCallPark", headers=self._headers)
            if r.status_code == 200:
                self._park_location = r.headers['Content-Location']
                self._park_address = self._park_location.split("?")[-1]
                self._park_extension = self._park_address.split(":")[-1]
                return self._park_extension
            else:
                raise NotAllowed("The call cannot be parked")
        else:
            params = {"address": extension}
            r = requests.put(self._url + f"/{self.id}/Park", headers=self._headers, params=params)
            if r.status_code == 200:
                self._park_location = r.headers['Content-Location']
                self._park_address = self._park_location.split("?")[-1]
                self._park_extension = self._park_address.split(":")[-1]
                return self._park_extension
            else:
                raise NotAllowed("The call cannot be parked")

    def reconnect(self):
        """Retrieves the call from hold **and releases all other calls**"""
        r = requests.put(self._url + f"{self.id}/Reconnect", headers=self._headers)
        self.held = False
        return XSIResponse(r)

    def recording(self, action: str):
        """Control the recording of the call

        For the recording() method to work, the user must have Call Recording enabled. Any unsuccessful attempt to
        record a call for a user who is not enabled for Call Recording will raise a NotAllowed exception.

        Args:
            action (str): The action to perform.

                'start': Starts recording, if it isn't in process already

                'stop': Stops the recording. Only applies to On Demand call recording.

                'pause': Pauses the recording

                'resume': Resume a paused recording

        Returns:
            bool: True if the recording command was accepted by the server

        Raises:
            wxcadm.exceptions.NotAllowed: The action is not allowed. Normally it indicates that the user does not have
                the Call Recording service assigned.
            ValueError: Raised when the action is not recognized.

        """
        log.info(f"Changing Recoding with an action: {action}")

        if action.lower() == "start":
            r = requests.put(self._url + f"/{self.id}/Record", headers=self._headers)
        elif action.lower() == "resume":
            r = requests.put(self._url + f"/{self.id}/ResumeRecording", headers=self._headers)
        elif action.lower() == "stop":
            r = requests.put(self._url + f"/{self.id}/StopRecording", headers=self._headers)
        elif action.lower() == "resume":
            r = requests.put(self._url + f"/{self.id}/PauseRecording", headers=self._headers)
        else:
            raise ValueError(f"{action} is not a valid action")

        log.debug(f"XSI Response:\n\t{r.text}")
        return XSIResponse(r)


class Conference:
    """The class for Conference Calls started by a Call.conference()"""

    def __init__(self, parent: object, calls: list, comment: str = ""):
        """Initialize a Conference instance for an XSI instance

        Args:
            parent (XSI): The XSI instance that owns this conference
            calls (list): Call IDs associated with the Conference. Always two Call IDs to start a Conference.
                Any additional Call IDs will be added to the conference as it is created.
            comment (str, optional): An optional text comment for the Conference

        Returns:
            Conference: This instance of the Conference class

        """
        self._parent: XSI = parent
        self._calls: list = calls
        self._userid = self._parent.id
        self._headers = self._parent._headers
        self._url = self._parent.xsi_endpoints['actions_endpoint'] + f"/v2.0/user/{self._userid}/calls/Conference/"
        self.comment: str = comment
        """Text comment associated with the Conference"""

    def deaf(self, call: str):
        """Stop audio and video from being sent to a participant. Audio and video from that participant are unaffected.

        Args:
            call (str): The Call ID to make deaf. The Call ID must be part of the Conference.

        Returns:
            bool: True if the command was successful

        Raises:
            wxcadm.exceptions.NotAllowed: Raised when the server rejects the command

        """
        r = requests.put(self._url + f"{call}/Deaf", headers=self._headers)
        return XSIResponse(r)

    def mute(self, call: str):
        """Mute a conference participant. Audio and video sent to the participant are unaffected.

        Args:
            call (str): The Call ID to mute. The Call ID must be part of the Conference

        Returns:
            bool: True if the command was successful

        Raises:
            wxcadm.exceptions.NotAllowed: Raised when the server rejects the command

        """
        r = requests.put(self._url + f"{call}/Mute", headers=self._headers)
        return XSIResponse(r)



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass


class TestECBN(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        # Enable for test debugging
        # wxcadm.console_logging()

    def test_person_ecbn(self):
        thing = choice(self.webex.org.people.webex_calling())
        ecbn = thing.ecbn
        self.assertIsInstance(ecbn, dict)
        prev_value = ecbn['selected']
        self.assertIsInstance(prev_value, str)
        with self.subTest("Location ECBN"):
            self.assertTrue(thing.set_ecbn('location'))
            self.assertTrue(thing.set_ecbn(prev_value))

    def test_workspace_ecbn(self):
        thing = choice(self.webex.org.workspaces.webex_calling())
        ecbn = thing.ecbn
        self.assertIsInstance(ecbn, dict)
        prev_value = ecbn['selected']
        self.assertIsInstance(prev_value, str)
        with self.subTest("Location ECBN"):
            self.assertTrue(thing.set_ecbn('location'))
            if prev_value.upper() != 'NONE':
                self.assertTrue(thing.set_ecbn(prev_value))

    def test_virtual_line_ecbn(self):
        thing = choice(self.webex.org.virtual_lines)
        ecbn = thing.ecbn
        self.assertIsInstance(ecbn, dict)
        prev_value = ecbn['selected']
        print(prev_value)
        self.assertIsInstance(prev_value, str)
        with self.subTest("Location ECBN"):
            self.assertTrue(thing.set_ecbn('location'))
            if prev_value.upper() != 'NONE':
                self.assertTrue(thing.set_ecbn(prev_value))


if __name__ == '__main__':
    unittest.main()


import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass


class TestRecordings(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        # Enable for test debugging
        # wxcadm.console_logging()

    def test_recording_list(self):
        with self.subTest("Org Recordings"):
            org_recordings = self.webex.org.get_recordings()
            self.assertIsInstance(org_recordings, wxcadm.RecordingList)
            if len(org_recordings) > 0:
                random_recording = choice(org_recordings)
                self.assertIsInstance(random_recording, wxcadm.Recording)
                del random_recording
            del org_recordings
        with self.subTest("Person Recordings"):
            person = choice(self.webex.org.people.webex_calling())
            person_recordings = self.webex.org.get_recordings(owner=person)
            self.assertIsInstance(person_recordings, wxcadm.RecordingList)
            if len(person_recordings) > 0:
                random_recording = choice(person_recordings)
                self.assertIsInstance(random_recording, wxcadm.Recording)
                del random_recording
            del person
            del person_recordings
        with self.subTest("Location Recordings"):
            location = choice(self.webex.org.locations.webex_calling())
            location_recordings = self.webex.org.get_recordings(location=location)
            self.assertIsInstance(location_recordings, wxcadm.RecordingList)
            if len(location_recordings) > 0:
                random_recording = choice(location_recordings)
                self.assertIsInstance(random_recording, wxcadm.Recording)
                del random_recording
            del location_recordings
            del location

    def test_recording_data(self) -> None:
        org_recordings = self.webex.org.get_recordings()
        if len(org_recordings) == 0:
            self.skipTest("No Recordings")
        recording = choice(org_recordings)
        self.assertIsInstance(recording.topic, str)
        self.assertIsInstance(recording.created, str)
        self.assertIsInstance(recording.details, dict)
        url = recording.url
        self.assertIsInstance(url, str)
        self.assertGreater(len(url), 10)


if __name__ == '__main__':
    unittest.main()


from __future__ import annotations

from typing import Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from wxcadm import Org, Location
from dataclasses import dataclass
from collections import UserList

import wxcadm.location
from wxcadm import log
from .common import *


class AutoAttendantList(UserList):
    def __init__(self, parent: Union[Org, Location]):
        super().__init__()
        log.debug("Initializing AutoAttendantLisa instance")
        self.parent: Union["Org", "Location"] = parent
        self.data: list = self._get_items()

    def _get_items(self):
        if isinstance(self.parent, wxcadm.Org):
            log.debug("Using Org ID as data filter")
            params = {'orgId': self.parent.id}
        elif isinstance(self.parent, wxcadm.Location):
            log.debug("Using Location as data filter")
            params = {'locationId': self.parent.id}
        else:
            raise ValueError("Unsupported parent class")

        log.debug("Getting Auto Attendant list")
        response = webex_api_call('get', f'v1/telephony/config/autoAttendants', params=params)
        log.debug(f"Received {len(response['autoAttendants'])} entries")
        items = []
        for entry in response['autoAttendants']:
            items.append(AutoAttendant(parent=self.parent, id=entry['id'], data=entry))
        return items

    def refresh(self):
        """ Refresh the list of Auto Attendants from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_items()
        return True

    def get(self, id: str = None, name: str = None, spark_id: str = None, uuid: str = None):
        """ Get the AutoAttendant instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Auto
        Attendants will be searched in order by ID, Name, UUID and Spark ID. If no arguments are provided, the method
        will raise an Exception.

        Args:
            id (str, optional): The AutoAttendant ID to find
            name (str, optional): The AutoAttendant Name to find
            spark_id (str, optional): The Spark ID to find
            uuid (str, optional): The UUID to find

        Returns:
            AutoAttendant: The AutoAttendant instance correlating to the given search argument.
                None is returned if no AutoAttendant is found.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None and uuid is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for aa in self.data:
                if aa.id == id:
                    return aa
        if name is not None:
            for aa in self.data:
                if aa.name == name:
                    return aa
        if uuid is not None:
            for aa in self.data:
                if aa.spark_id.split('/')[-1].upper() == uuid.upper():
                    return aa
        if spark_id is not None:
            for aa in self.data:
                if aa.spark_id == spark_id:
                    return aa
        return None

    def create(self,
               name: str,
               first_name: str,
               last_name: str,
               phone_number: Optional[str],
               extension: Optional[str],
               business_hours_schedule: str,
               holiday_schedule: Optional[str],
               business_hours_menu: Optional[dict] = None,
               after_hours_menu: Optional[dict] = None,
               extension_dialing_scope: str = "GROUP",
               name_dialing_scope: str = "GROUP",
               language: Optional[str] = None,
               time_zone: Optional[str] = None,
               location: Optional[wxcadm.Location] = None
               ):
        """ Create a new Auto Attendant

        Args:
            name (str): The name of the Auto Attendant
            first_name (str): The first name of the Auto Attendant in the directory
            last_name (str): The last name of the Auto Attendant in the directory
            phone_number (str, optional): The phone number of the Auto Attendant
            extension (str, optional): The extension of the Auto Attendant
            business_hours_schedule (str): The name of the Business Hours Schedule to appy to the Auto Attendant
            holiday_schedule (str, optional): The name of the Holiday Schedule to apply to the Auto Attendant
            business_hours_menu (dict, optional): The Business Hours menu configuration. See the Webex Developer docs
                for details. Defaults to a menu with a zero (0) to exit the menu.
            after_hours_menu (dict, optional): The After Hours menu configuration. See the Webex Developer docs for
                details. Defaults to a menu with 1 zero (0) to exit the menu.
            extension_dialing_scope (str, optional): Whether extension dialing matches at the `'GROUP'` (i.e. Location)
                or `'ENTERPRISE'`. Defaults to `'GROUP'`
            name_dialing_scope (str, optional): Whether name dialing matches at the `'GROUP'` (i.e. Location) or
                `'ENTERPRISE'`. Defaults to `'GROUP'`
            language (str, optional): The language code for the Auto Attendant. Defaults to the Location language
            time_zone (str, optional): The time zone for the Auto Attendant. Defaults to the Location time zone
            location (Location, optional): The Location to build the Auto Attendant at. If the
                :class:`AutoAttendantList` is at the Location level, this parameter is optional. If it is at the
                Org level, it is required.

        Returns:
            AutoAttendant: The :class:`AutoAttendant` that is created.

        """
        if location is None and isinstance(self.parent, wxcadm.Org):
            raise ValueError("location is required for Org-level AutoAttendantList")
        elif location is None and isinstance(self.parent, wxcadm.Location):
            location = self.parent
        log.info(f"Creating Auto Attendant at Location {location.name} with name: {name}")
        if location.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        # Get some values if they weren't passed
        if phone_number is None and extension is None:
            raise ValueError("phone_number and/or extension are required")
        if time_zone is None:
            log.debug(f"Using Location time_zone {location.time_zone}")
            time_zone = location.time_zone
        if language is None:
            log.debug(f"Using Location announcement_language {location.announcement_language}")
            language = location.announcement_language
        if business_hours_menu is None or after_hours_menu is None:
            default_menu = {
                'greeting': 'DEFAULT',
                'extensionEnabled': True,
                'keyConfigurations': [
                    {'key': '0', 'action': 'EXIT'}
                ]
            }
            if business_hours_menu is None:
                log.debug('Using default menu for Business Hours')
                business_hours_menu = default_menu
            if after_hours_menu is None:
                log.debug('Using default menu for After Hours')
                after_hours_menu = default_menu

        payload = {
            "name": name,
            "firstName": first_name,
            "lastName": last_name,
            "extension": extension,
            "phoneNumber": phone_number,
            "timeZone": time_zone,
            "languageCode": language,
            "businessSchedule": business_hours_schedule,
            "holidaySchedule": holiday_schedule,
            "extensionDialing": extension_dialing_scope,
            "nameDialing": name_dialing_scope,
            "businessHoursMenu": business_hours_menu,
            "afterHoursMenu": after_hours_menu
        }
        response = webex_api_call("post", f"v1/telephony/config/locations/{location.id}/autoAttendants",
                                  payload=payload)
        new_aa_id = response['id']
        self.refresh()
        return self.get(id=new_aa_id)


@dataclass
class AutoAttendant:
    parent: Union[wxcadm.Org, wxcadm.Location]
    """ The Org instance that ows this AutoAttendant """
    id: str
    """ The Webex ID of the AutoAttendant """
    data: dict

    def __post_init__(self):
        # Attributes available in the main list
        self.name = self.data.get('name', '')
        self.location_id = self.data.get('locationId', '')
        self.phone_number: Optional[str] = self.data.get('phoneNumber', None)
        self.extension: Optional[str] = self.data.get('extension', None)
        self.routing_prefix: Optional[str] = self.data.get('routingPrefix', None)
        self.esn: Optional[str] = self.data.get('esn', None)
        self.toll_free_number: bool = self.data.get('tollFreeNumber', False)
        # Attributes that require a Get Detail call
        self._first_name: Optional[str] = None
        self._last_name: Optional[str] = None
        self._enabled: Optional[bool] = None
        self._alternate_numbers: Optional[list] = None
        self._language: Optional[str] = None
        self._language_code: Optional[str] = None
        self._business_schedule: Optional[str] = None
        self._holiday_schedule: Optional[str] = None
        self._extension_dialing: Optional[str] = None
        self._name_dialing: Optional[str] = None
        self._time_zone: Optional[str] = None
        self._business_hours_menu: Optional[dict] = None
        self._after_hours_menu: Optional[dict] = None

    @property
    def enabled(self) -> bool:
        if self._enabled is None:
            self._get_details()
        return self._enabled

    @property
    def first_name(self) -> str:
        if self._first_name is None:
            self._get_details()
        return self._first_name

    @property
    def last_name(self) -> str:
        if self._last_name is None:
            self._get_details()
        return self._last_name

    @property
    def alternate_numbers(self) -> list:
        if self._alternate_numbers is None:
            self._get_details()
        return self._alternate_numbers

    @property
    def language(self) -> str:
        if self._language is None:
            self._get_details()
        return self._language

    @property
    def language_code(self) -> str:
        if self._language_code is None:
            self._get_details()
        return self._language_code

    @property
    def business_schedule(self) -> str:
        if self._business_schedule is None:
            self._get_details()
        return self._business_schedule

    @property
    def holiday_schedule(self) -> str:
        if self._holiday_schedule is None:
            self._get_details()
        return self._holiday_schedule

    @property
    def extension_dialing(self) -> str:
        if self._extension_dialing is None:
            self._get_details()
        return self._extension_dialing

    @property
    def name_dialing(self) -> str:
        if self._name_dialing is None:
            self._get_details()
        return self._name_dialing

    @property
    def time_zone(self) -> str:
        if self._time_zone is None:
            self._get_details()
        return self._time_zone

    @property
    def business_hours_menu(self) -> dict:
        if self._business_hours_menu is None:
            self._get_details()
        return self._business_hours_menu

    @property
    def after_hours_menu(self) -> dict:
        if self._after_hours_menu is None:
            self._get_details()
        return self._after_hours_menu

    @property
    def spark_id(self) -> str:
        return decode_spark_id(self.id)

    @property
    def config(self) -> dict:
        """ The JSON config of the AutoAttendant """
        response = self._get_details()
        return response

    def _get_details(self):
        response = webex_api_call("get", f"v1/telephony/config/locations/{self.location_id}/autoAttendants/{self.id}")
        self._enabled = response.get('enabled', None)
        self._first_name = response.get('firstName', '')
        self._last_name = response.get('lastName', '')
        self._alternate_numbers = response.get('alternateNumbers', [])
        self._language = response.get('language', '')
        self._language_code = response.get('languageCode', '')
        self._business_schedule = response.get('businessSchedule', '')
        self._holiday_schedule = response.get('holidaySchedule', None)
        self._extension_dialing = response.get('extensionDialing', '')
        self._name_dialing = response.get('nameDialing', '')
        self._time_zone = response.get('timeZone', '')
        self._business_hours_menu = response.get('businessHoursMenu', {})
        self._after_hours_menu = response.get('afterHoursMenu', {})
        return response

    @property
    def call_forwarding(self) -> dict:
        """ The Call Forwarding settings for the AutoAttendant """
        response = webex_api_call("get", f"v1/telephony/config/locations/{self.location_id}/autoAttendants/{self.id}"
                                         f"/callForwarding")
        return response

    def copy_menu_from_template(self, source: AutoAttendant, menu_type: str = "both"):
        """ Copy the Business Hours, After Hours, or both menus from another :class:`AutoAttendant` instance.

        Note that this copy functionality will not work (yet) if the source Auto Attendant uses a CUSTOM announcement,
        which, unfortunately, is most of them. The Webex API should support this soon, but, until them, the method's
        usefulness is limited.

        Args:
            source (AutoAttendant): The source :class:`AutoAttendant` instance to copy from
            menu_type (str, optional): The menu to copy. 'business_hours', 'after_hours' or 'both' are valid. Defaults
                to 'both'.

        Returns:
            bool: True on success. False otherwise.

        Raises:
            ValueError: Raised when the menu_type value is not one of the accepted values.

        """
        if menu_type.lower() == "both":
            self.config['businessHoursMenu'] = source.config['businessHoursMenu']
            self.config['afterHoursMenu'] = source.config['afterHoursMenu']
        elif menu_type.lower() == "business_hours":
            self.config['businessHoursMenu'] = source.config['businessHoursMenu']
        elif menu_type.lower() == "after_hours":
            self.config['afterHoursMenu'] = source.config['afterHoursMenu']
        else:
            raise ValueError(f"{menu_type} must be 'business_hours', 'after_hours' or 'both'")

        resp = webex_api_call("put", f"v1/telephony/config/locations/{self.location_id}/autoAttendants/{self.id}",
                              payload=self.config)
        if resp:
            return True
        else:
            return False



from __future__ import annotations

import base64
import logging
import uuid
import time
import re
import requests
import sys
from typing import Optional, TYPE_CHECKING

import wxcadm

if TYPE_CHECKING:
    from requests_toolbelt import MultipartEncoder

from .exceptions import *
import wxcadm
from wxcadm import log

__all__ = ['decode_spark_id', 'console_logging', 'tracking_id', 'webex_api_call', '_url_base', '_webex_headers',
           'location_finder']

# Some functions available to all classes and instances (optionally)
_url_base = "https://webexapis.com/"
_webex_headers = {"Authorization": "",
                  "Content-Type": "application/json",
                  "Accept": "application/json"}


def webex_api_call(method: str,
                   url: str,
                   headers: Optional[dict] = None,
                   params: Optional[dict] = None,
                   payload: dict | MultipartEncoder | None = None,
                   retry_count: Optional[int] = 5,
                   domain: Optional[str] = None,
                   **kwargs):
    """ Generic handler for all Webex API requests

    This function performs the Webex API call as a Session and handles processing the response. It has the ability
    to recognize paginated responses from the API and make subsequent requests to get all data, regardless of
    how many pages (calls) are needed.

    Args:
        method (str): The HTTP method to use. **get**, **post** and **put** are supported.
        url (str): The endpoint part of the URL (after https://webexapis.com/)
        headers (dict, optional): HTTP headers to use with the request. If not provided, **wxcadm** will use the base
            Authorization header from when the Webex instance was initialized.
        params (dict, optional): Any parameters to be passed as part of an API call
        payload (dict, optional): Payload that will be sent in a POST or PUT. Will be converted to JSON during the
            API call
        retry_count (int, optional): Controls the number of times an API call will be retried if the API returns a
            429 Too Many Requests. The wait time between retries will be based on the Retry-After header sent by Webex.
            Default is 5.
        domain (str, optional): The domain name to use if anything other than https://webexapis.com

    Returns:
        The return value will vary based on the API response. If a list of items are returned, a list will be returned.
            If the details for a single entry are returned by the API, a dict will be returned.

    Raises:
        wxcadm.exceptions.APIError: Raised when the API call fails to retrieve at least one response.

    """
    log.debug("Webex API Call:")
    log.debug(f"\tMethod: {method}")

    # Hacky fix for API calls that don't use the webexapis.com base domain
    if domain is not None:
        url_base = domain
    else:
        url_base = _url_base

    log.debug(f"\tURL: {url_base + url}")
    log.debug(f"\tParams: {params}")

    start = time.time()     # Tracking API execution time
    session = requests.Session()
    if headers is not None:
        session.headers.update(headers)
    else:
        session.headers.update(_webex_headers)

    try_num = 1
    while try_num <= retry_count:
        if method.lower() == "get":
            r = session.get(url_base + url, params=params)
            if r.ok:
                response = r.json()
                # With an 'items' array, we know we are getting multiple values.
                # Without it, we are getting a singe entity
                if "items" in response:
                    log.debug(f"Webex returned {len(response['items'])} items")
                else:
                    session.close()
                    return response
            else:
                log.warning("Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                log.warning(f"\t[{r.status_code}] {r.text}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                if r.status_code == 400 and kwargs.get('ignore_400', False) is True:
                    log.info("Ignoring 400 Error due to ignore_400=True")
                    session.close()
                    return None
                # The following was added to handle cross-region analytics and CDR
                if r.status_code == 451:
                    log.info("Retrying GET in different API region")
                    message = r.json()
                    log.debug(message['message'])
                    m = re.search('Please use (.*)', message['message'])
                    if m:
                        new_domain = m.group(1)
                        log.info(f'Using {new_domain} as new domain')
                        url_base = f'https://{new_domain}'
                        continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)

            # Now we look for pagination and get any additional pages as part of the same Session
            if "next" in r.links:
                keep_going = True
                next_url = r.links['next']['url']
                while keep_going:
                    log.debug(f"Getting more items from {next_url}")
                    r = session.get(next_url)
                    if r.ok:
                        new_items = r.json()
                        if "items" not in new_items:
                            continue     # This is here just to handle a weird case where the API responded with no data
                        log.debug(f"Webex returned {len(new_items['items'])} more items")
                        response['items'].extend(new_items['items'])
                        if "next" not in r.links:
                            keep_going = False
                        else:
                            next_url = r.links['next']['url']
                    else:
                        if r.status_code == 429:
                            retry_after = int(r.headers.get('Retry-After', 30))
                            log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                            time.sleep(retry_after)
                            continue
                        else:
                            keep_going = False

            session.close()
            end = time.time()
            log.debug(f"__webex_api_call() completed in {end - start} seconds")
            return response['items']
        elif method.lower() == "put":
            log.debug(f"\tPayload: {payload}")
            r = session.put(url_base + url, params=params, json=payload)
            if r.ok:
                try:
                    response = r.json()
                except requests.exceptions.JSONDecodeError:
                    response = r.text
                if response:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
            else:
                log.warning("Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        elif method.lower() == "put_upload":
            log.debug("Putting a file upload")
            log.debug(payload)
            session.headers['Content-Type'] = payload.content_type
            r = session.put(url_base + url, params=params, data=payload)
            log.debug(f"Response Headers: {r.headers}")
            if r.ok:
                try:
                    response = r.json()
                except requests.exceptions.JSONDecodeError:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
            else:
                log.warning("Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        elif method.lower() == "post":
            log.debug(f"Post body: {payload}")
            r = session.post(url_base + url, params=params, json=payload)
            if r.ok:
                try:
                    response = r.json()
                    log.debug(f"Response: {response}")
                except requests.exceptions.JSONDecodeError:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
            else:
                log.warning(f"Webex API returned an error: {r.text}")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        elif method.lower() == "post_upload":
            log.debug("Posting a file upload")
            log.debug(payload)
            session.headers['Content-Type'] = payload.content_type
            r = session.post(url_base + url, params=params, data=payload)
            log.debug(f"Response Headers: {r.headers}")
            if r.ok:
                try:
                    response = r.json()
                except requests.exceptions.JSONDecodeError:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
            else:
                log.warning("Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        elif method.lower() == "delete":
            log.debug(f"Post body: {payload}")
            r = session.delete(url_base + url, params=params)
            if r.ok:
                try:
                    response = r.json()
                    log.debug(f'Response: {response}')
                except requests.exceptions.JSONDecodeError:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
            else:
                log.warning("Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        elif method.lower() == "patch":
            r = session.patch(url_base + url, params=params, json=payload)
            if r.ok:
                try:
                    response = r.json()
                except requests.exceptions.JSONDecodeError:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return True
                else:
                    end = time.time()
                    log.debug(f"__webex_api_call() completed in {end - start} seconds")
                    session.close()
                    return response
            else:
                log.info(f"Webex API returned an error")
                log.debug(f"TrackingID: {r.headers.get('Trackingid', 'None')}")
                if r.status_code == 429:
                    retry_after = int(r.headers.get('Retry-After', 30))
                    log.info(f"Received 429 Too Many Requests. Waiting {retry_after} seconds to retry.")
                    time.sleep(retry_after)
                    continue
                else:
                    session.close()
                    try:
                        raise APIError(r.json())
                    except requests.exceptions.JSONDecodeError:
                        raise APIError(r.text)
        else:
            return False
    return False


def console_logging(level: str = "debug", formatter: Optional[logging.Formatter] = None):
    """ Enable logging directly to STDOUT

    This adds a STDOUT logging handler to the existing logger. Any other handlers that have been applied will continue
    to function at their current logging level.

    Args:
        level (str, optional): The logging level. Valid values are ``"none"``, ``"debug"``, ``"info"``, ``"warning"``
            and ``"critical"``. Defaults to ``debug`` for full debug output.
        formatter (logging.Formatter, optional): A :py:class:`logging.Formatter` object defining the log format for
            logging to the console. If omitted, the format will be the message level and the message.
    """
    handler = logging.StreamHandler(sys.stdout)
    level_map = {"info": logging.INFO,
                 "warning": logging.WARNING,
                 "debug": logging.DEBUG,
                 "critical": logging.CRITICAL}
    if level.lower() != "none":
        handler.setLevel(level_map[level])
        if formatter is not None:
            if isinstance(formatter, logging.Formatter):
                handler.setFormatter(formatter)
            else:
                log.warning("Did not receive logging.Formatter object for formatter argument")
                formatter = logging.Formatter('%(levelname)s:\t%(message)s')
                handler.setFormatter(formatter)
        else:
            formatter = logging.Formatter('%(levelname)s:\t%(message)s')
            handler.setFormatter(formatter)
        log.addHandler(handler)
    else:
        log.debug("Removing Console Logging handlers")
        for handler in log.handlers:
            if isinstance(handler, logging.StreamHandler) and handler.stream.name == "<stdout>":
                log.removeHandler(handler)


def decode_spark_id(id: str):
    """ Decode the Webex ID to obtain the Spark ID

    Note that the returned string is the full URI, like
        ```ciscospark://us/PEOPLE/5b7ddefe-cc47-496a-8df0-18d8e4182a99```. In most cases, you only care about the ID
        at the end, so a ```.split('/')[-1]``` can be used to obtain that.

    Args:
        id (str): The Webex ID (base64 encoded string)

    Returns:
        str: The Spark ID

    """
    id_bytes = base64.b64decode(id + "==")
    spark_id: str = id_bytes.decode("utf-8")
    return spark_id


def tracking_id():
    id = f"WXCADM_{uuid.uuid4()}"
    return id


def location_finder(location_id: str, parent):
    """ Return the :class:`Location` for a given Location ID """
    log.debug(f"location_finder: Finding {location_id}")
    if isinstance(parent, wxcadm.Location):
        log.debug("location_finder: Parent is a Location")
        if parent.id == location_id:
            log.debug(f"location_finder: Parent ID matches ({parent.name})")
            return parent
        else:
            log.debug("location_finder: Parent ID doesn't match. Finding by ID at Org.")
            location = parent.parent.locations.get(id=location_id)
            if location is None:
                log.debug("No match found. Returning None.")
            else:
                log.debug(f"Location found ({location.name})")
            return location
    elif isinstance(parent, wxcadm.Org):
        log.debug("location_finder: Parent is an Org")
        location = parent.locations.get(id=location_id)
        if location is None:
            log.debug("No match found. Returning None.")
        else:
            log.debug(f"Location found ({location.name})")
        return location
